from euphorie.client import model
from euphorie.client.browser.session import Profile
from euphorie.client.profile import AddToTree
from euphorie.client.profile import BuildSurveyTree
from euphorie.client.profile import extractProfile
from euphorie.client.tests.test_update import TreeTests
from euphorie.content.profilequestion import IProfileQuestion
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from unittest import mock
from z3c.saconfig import Session
from zope.interface import alsoProvides


def createContainer(id, profile=False):
    from euphorie.content.interfaces import IQuestionContainer
    from zope.interface import implementer

    @implementer(IQuestionContainer)
    class Container(dict):
        title = "container"
        description = None

        def __init__(self, id):
            self.id = id

    c = Container(id)
    if profile:
        alsoProvides(c, IProfileQuestion)
    return c


def createRisk(id):
    from euphorie.content.risk import IRisk
    from zope.interface import implementer

    @implementer(IRisk)
    class Risk:
        title = "risk"
        type = "risk"
        default_probability = 0
        default_frequency = 0
        default_effect = 0
        description = None
        evaluation_method = "calculated"
        risk_always_present = None

        def __init__(self, id):
            self.id = id

    return Risk(id)


def createModule(id):
    from euphorie.content.interfaces import IQuestionContainer
    from euphorie.content.module import IModule
    from zope.interface import implementer

    @implementer(IModule, IQuestionContainer)
    class Module(dict):
        title = "module"
        description = None
        optional = False
        solution_direction = False

        def __init__(self, id):
            dict.__init__(self)
            self.id = id

    return Module(id)


class AddToTreeTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.session = Session()
        account = model.Account(id=1, loginname="jane", password="secret")
        self.session.add(account)
        self.survey = model.SurveySession(
            title="Survey", zodb_path="survey", account=account
        )
        self.session.add(self.survey)
        self.session.flush()
        self.root = self.survey.addChild(
            model.Module(title="test session", module_id="1", zodb_path="1")
        )

    def testAddNonInterestingNode(self):
        AddToTree(self.root, None, title="title")
        self.assertEqual(list(self.root.children()), [])

    def testAddRisk(self):
        question = createRisk("13")
        AddToTree(self.root, question)
        children = list(self.root.children())
        self.assertEqual(len(children), 1)

        child = children[0]
        self.assertTrue(isinstance(child, model.Risk))
        self.assertEqual(child.title, "risk")
        self.assertEqual(child.risk_id, "13")
        self.assertEqual(child.risk_type, "risk")

    def testAddPolicy(self):
        question = createRisk("13")
        question.type = "policy"
        AddToTree(self.root, question)
        children = list(self.root.children())
        child = children[0]
        self.assertEqual(child.risk_type, "policy")
        self.assertEqual(child.priority, "high")

    def testAddTop5Risk(self):
        question = createRisk("13")
        question.type = "top5"
        AddToTree(self.root, question)
        children = list(self.root.children())
        child = children[0]
        self.assertEqual(child.risk_type, "top5")
        self.assertEqual(child.priority, "high")

    def testAddEmptyContainer(self):
        container = createContainer("13")
        AddToTree(self.root, container)
        children = list(self.root.children())
        self.assertEqual(len(children), 1)
        child = children[0]
        self.assertTrue(isinstance(child, model.Module))
        self.assertEqual(child.title, "container")
        self.assertEqual(child.module_id, "13")

    def testOverrideTitle(self):
        question = createRisk("13")
        AddToTree(self.root, question, title="other title")
        self.assertEqual(self.root.children().first().title, "other title")

    def testContainerWithChildren(self):
        container = createContainer("13")
        container["5"] = createRisk("5")
        container["11"] = createRisk("11")
        AddToTree(self.root, container)
        children = list(self.root.children())
        self.assertEqual(len(children), 1)
        child = children[0]
        grandchildren = list(child.children())
        self.assertEqual(len(grandchildren), 2)

    def testContainerWithTitleAndProfile(self):
        # This is a test for #96
        container = createContainer("2")
        container["2"] = createContainer("3")
        AddToTree(self.root, container, [], title="Top level title", profile_index=1)
        children = list(self.root.children())
        self.assertEqual(len(children), 1)
        child = children[0]
        self.assertEqual(child.title, "Top level title")
        # Profile index is forced back to 0 since root has profile index 0
        self.assertEqual(child.profile_index, 0)

    def test_item_with_description(self):
        question = createRisk("13")
        question.description = "<p>Hello!</p>"
        AddToTree(self.root, question)
        children = list(self.root.children())
        child = children[0]
        self.assertEqual(child.has_description, True)

    def test_item_without_description(self):
        question = createRisk("13")
        question.description = "<p></p>"
        AddToTree(self.root, question)
        children = list(self.root.children())
        child = children[0]
        self.assertEqual(child.has_description, False)

    def test_pretend_optional_module_has_description(self):
        # This is necessary to make sure optional modules are never skipped.
        module = createModule("13")
        module.optional = True
        AddToTree(self.root, module)
        children = list(self.root.children())
        child = children[0]
        self.assertEqual(child.has_description, True)


class BuildSurveyTreeTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.session = Session()
        account = model.Account(id=1, loginname="jane", password="secret")
        self.session.add(account)
        self.survey = model.SurveySession(
            title="Survey", zodb_path="survey", account=account
        )
        self.session.add(self.survey)
        self.session.flush()

    def test_empty_profile_no_question(self):
        BuildSurveyTree({}, {}, dbsession=self.survey)
        self.assertTrue(not self.survey.hasTree())

    def test_empty_profile_with_risk(self):
        BuildSurveyTree({"one": createRisk("13")}, {}, dbsession=self.survey)
        self.assertTrue(self.survey.hasTree())
        children = self.survey.children().all()
        self.assertEqual(len(children), 1)
        self.assertEqual(children[0].zodb_path, "13")
        self.assertEqual(children[0].children().count(), 0)

    def test_empty_profile_with_container(self):
        BuildSurveyTree({"one": createContainer("13")}, {}, dbsession=self.survey)
        self.assertTrue(self.survey.hasTree())
        children = self.survey.children().all()
        self.assertEqual(len(children), 1)
        self.assertEqual(children[0].zodb_path, "13")
        self.assertEqual(children[0].children().count(), 0)

    def test_empty_profile_with_profile_question(self):
        BuildSurveyTree({"one": createContainer("13", True)}, {}, dbsession=self.survey)
        self.assertTrue(not self.survey.hasTree())

    def testOptionalProfilePositive(self):
        BuildSurveyTree(
            {"one": createContainer("13", True)},
            profile={"13": True},
            dbsession=self.survey,
        )
        self.assertTrue(self.survey.hasTree())

    def testOptionalProfileNegative(self):
        BuildSurveyTree(
            {"one": createContainer("13", True)},
            profile={"13": False},
            dbsession=self.survey,
        )
        self.assertTrue(not self.survey.hasTree())

    def test_profile_without_answers(self):
        BuildSurveyTree(
            {"one": createContainer("13", True)},
            profile={"13": []},
            dbsession=self.survey,
        )
        self.assertTrue(not self.survey.hasTree())

    def test_profile_with_multiple_answers(self):
        BuildSurveyTree(
            {"one": createContainer("13", True)},
            profile={"13": ["one", "two"]},
            dbsession=self.survey,
        )
        self.assertTrue(self.survey.hasTree())
        children = self.survey.children().all()
        self.assertEqual(len(children), 1)
        self.assertEqual(children[0].zodb_path, "13")
        self.assertEqual(children[0].profile_index, -1)
        grandchildren = children[0].children().all()
        self.assertEqual(len(grandchildren), 2)
        self.assertEqual(grandchildren[0].title, "one")
        self.assertEqual(grandchildren[0].profile_index, 0)
        self.assertEqual(grandchildren[1].title, "two")
        self.assertEqual(grandchildren[1].profile_index, 1)


class ExtractProfileTests(TreeTests):
    def extractProfile(self, *a, **kw):
        return extractProfile(*a, **kw)

    def testNoProfileQuestions(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        session = self.createSurveySession()
        session.addChild(model.Module(title="Root", module_id="1", zodb_path="1"))
        self.assertEqual(self.extractProfile(survey, session), {})

    def testOptionalProfileQuestionNotSelected(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        pq = survey["1"]
        pq.use_location_question = False
        session = self.createSurveySession()
        self.assertEqual(self.extractProfile(survey, session), {"1": False})

    def testOptionalProfileQuestionSelected(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        pq = survey["1"]
        pq.use_location_question = False
        session = self.createSurveySession()
        session.addChild(model.Module(title="Root", module_id="1", zodb_path="1"))
        self.assertEqual(self.extractProfile(survey, session), {"1": True})

    def testRepeatableProfileQuestionNotSelected(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        session = self.createSurveySession()
        self.assertEqual(self.extractProfile(survey, session), {"1": []})

    def testRepeatableProfileQuestionSingleOption(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        pq = survey["1"]
        pq.title = "Repeatable profile question"
        session = self.createSurveySession()
        session.addChild(
            model.Module(title="First answer", module_id="1", zodb_path="1")
        )
        self.assertEqual(self.extractProfile(survey, session), {"1": ["First answer"]})


class Profile_getDesiredProfile_Tests(TreeTests):
    def getDesiredProfile(self, survey, form):
        class Request:
            def __init__(self, form):
                self.form = form

            @property
            def request(self):
                return self

        view = Profile(survey, Request(form))
        return view.getDesiredProfile()

    def testNoProfile(self):
        survey = self.createClientSurvey()
        self.assertEqual(self.getDesiredProfile(survey, {}), {})

    def testNoProfile_FluffInForm(self):
        survey = self.createClientSurvey()
        self.assertEqual(self.getDesiredProfile(survey, {"dummy": True}), {})

    def testOptionalProfile_NoAnswer(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        pq = survey["1"]
        pq.use_location_question = False
        self.assertEqual(self.getDesiredProfile(survey, {}), {})

    def testOptionalProfile_FalseAnswer(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        pq = survey["1"]
        pq.use_location_question = False
        self.assertEqual(
            self.getDesiredProfile(survey, {"pq1.present": False}), {"1": False}
        )

    def testOptionalProfile_NegativeAnswer(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        pq = survey["1"]
        pq.use_location_question = False
        self.assertEqual(
            self.getDesiredProfile(survey, {"pq1.present": "no"}), {"1": False}
        )

    def testOptionalProfile_TrueAnswer(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        pq = survey["1"]
        pq.use_location_question = False
        self.assertEqual(
            self.getDesiredProfile(survey, {"pq1.present": "yes"}), {"1": True}
        )

    def test_profile_no_anwer(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        self.assertEqual(self.getDesiredProfile(survey, {}), {})

    def test_profile_whitespace_answer(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        self.assertEqual(
            self.getDesiredProfile(survey, {"1": ["   "], "pq1.present": "yes"}),
            {"1": []},
        )

    def test_profile_with_two_answers_not_multiple_active(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        self.assertEqual(
            self.getDesiredProfile(survey, {"1": ["one", "two"], "pq1.present": "yes"}),
            {"1": ["one"]},
        )

    def test_profile_with_two_answers_multiple_active(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        self.assertEqual(
            self.getDesiredProfile(
                survey,
                {"1": ["one", "two"], "pq1.present": "yes", "pq1.multiple": "yes"},
            ),
            {"1": ["one", "two"]},
        )

    def test_profile_strip_whitespace(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        self.assertEqual(
            self.getDesiredProfile(survey, {"1": [" *** "], "pq1.present": "yes"}),
            {"1": ["***"]},
        )


class Profile_setupSession_Tests(TreeTests):
    def makeView(self, survey, session=None):
        if not session:
            session = self.createSurveySession()
        return api.content.get_view(
            "profile",
            survey.restrictedTraverse("++session++%s" % session.id),
            self.get_client_request(),
        )

    def test_NewSession_NoProfile(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        mod = survey["1"]
        mod.title = "Module one"
        view = self.makeView(survey)

        session = view.session
        with api.env.adopt_user(user=session.account):
            view.setupSession()
            self.assertTrue(view.session is session)
            self.assertEqual(session.hasTree(), True)

    def test_NewSession_SimpleProfile(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        view = self.makeView(survey)
        session = view.session
        view.request.form["1"] = ["one"]
        view.request.form["pq1.present"] = "yes"
        with api.env.adopt_user(user=session.account):
            view.setupSession()
            self.assertEqual(view.session, session)
            self.assertEqual(session.hasTree(), True)

    def test_NewSession_EmptyProfile(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        view = self.makeView(survey)
        session = view.session
        with api.env.adopt_user(user=session.account):
            view.setupSession()
            self.assertTrue(view.session is session)
            self.assertEqual(session.hasTree(), False)

    def test_ExistingSession_NoProfile(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        mod = survey["1"]
        mod.title = "Module one"
        view = self.makeView(survey)
        with mock.patch("euphorie.client.profile.extractProfile", return_value={}):
            session = view.session
            BuildSurveyTree(survey, {}, session)
            with api.env.adopt_user(user=session.account):
                view.setupSession()
                self.assertEqual(view.session, session)
                self.assertTrue(session.hasTree())

    def test_ExistingSession_NoProfileChange(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        survey.invokeFactory("euphorie.profilequestion", "2")
        survey.get("2").use_location_question = False

        view = self.makeView(survey)
        with mock.patch(
            "euphorie.client.profile.extractProfile",
            return_value={"1": ["London"], "2": True},
        ):
            session = view.session
            BuildSurveyTree(survey, {"1": ["London"], "2": True}, session)
            view.request.form["1"] = ["London"]
            view.request.form["pq1.present"] = "yes"
            view.request.form["pq2.present"] = "yes"
            with api.env.adopt_user(user=session.account):
                view.setupSession()
                self.assertEqual(view.session, session)
                self.assertEqual(view.session.hasTree(), True)

    def test_ExistingSession_ProfileChanged(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        view = self.makeView(survey)
        with mock.patch(
            "euphorie.client.profile.extractProfile",
            return_value={"1": ["London"]},
        ):
            session = view.session
            with api.env.adopt_user(user=session.account):
                BuildSurveyTree(survey, {"1": ["London"]}, session)
                view.request.form["1"] = ["London", "Paris"]
                # this will run an SQL statement incompatible with sqlite
                with mock.patch(
                    "euphorie.client.model.SurveySession.copySessionData",
                    return_value=None,
                ):
                    new_session = view.setupSession()
                view = self.makeView(survey, new_session)
                self.assertTrue(view.session is not session)
                self.assertEqual(view.session.hasTree(), False)

    def test_ExistingSession_ProfileChangedNoLocation(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        survey.get("1").use_location_question = False
        view = self.makeView(survey)
        with mock.patch(
            "euphorie.client.profile.extractProfile",
            return_value={"1": True},
        ):
            session = view.session
            with api.env.adopt_user(user=session.account):
                BuildSurveyTree(survey, {"1": True}, session)
                view.request.form["pq1.present"] = "no"
                # this will run an SQL statement incompatible with sqlite
                with mock.patch(
                    "euphorie.client.model.SurveySession.copySessionData",
                    return_value=None,
                ):
                    new_session = view.setupSession()
                view = self.makeView(survey, new_session)
                self.assertTrue(view.session is not session)
                self.assertEqual(view.session.hasTree(), False)

    def test_create_survey_session(self):
        survey = self.createClientSurvey()
        view = self.makeView(survey)
        survey_view = api.content.get_view("index_html", survey, view.request)
        new_session = survey_view.create_survey_session(
            "a",
            view.session.account,
            report_comment="b",
        )
        self.assertEqual(new_session.title, "a")
        self.assertEqual(new_session.account, view.session.account)
        self.assertEqual(new_session.zodb_path, view.session.zodb_path)
        self.assertEqual(new_session.report_comment, "b")



from euphorie.deployment.upgrade.utils import alembic_upgrade_to
from ftw.upgrade import UpgradeStep


class RemoveTheCompletionPercentageColumn(UpgradeStep):
    """Remove the completion_percentage column."""

    def __call__(self):
        alembic_upgrade_to(self.target_version)



from euphorie.content.tests.utils import _create
from euphorie.testing import EuphorieIntegrationTestCase
from plone.dexterity.interfaces import IDexterityFTI
from z3c.form.interfaces import IDataManager
from zope import component

import bcrypt


class CountryManagerTests(EuphorieIntegrationTestCase):
    def createCountryManager(self):
        manager = _create(self.portal.sectors["nl"], "euphorie.countrymanager", "mgr")
        return manager

    def testCanNotBeCopied(self):
        self.loginAsPortalOwner()
        manager = self.createCountryManager()
        self.assertFalse(manager.cb_isCopyable())

    def testPasswordsAreHashed(self):
        self.loginAsPortalOwner()
        manager = self.createCountryManager()

        fti = component.queryUtility(IDexterityFTI, name=manager.portal_type)
        field = fti.lookupSchema().get("password")
        component.getMultiAdapter((manager, field), IDataManager).set("secret")

        self.assertFalse(manager.password == "secret")
        self.assertTrue(bcrypt.hashpw("secret", manager.password) == manager.password)



from Acquisition import aq_parent
from z3c.form.datamanager import AttributeField
from zope.security.checker import canAccess
from zope.security.checker import canWrite
from zope.security.checker import Proxy


class ParentAttributeField(AttributeField):
    parent_mapping = {}

    @property
    def adapted_context(self):
        return self.context

    def _name_and_context(self):
        name = self.field.__name__
        context = self.adapted_context
        if name in self.parent_mapping:
            context = aq_parent(context)
            name = self.parent_mapping[name]
        return (name, context)

    def get(self):
        (name, context) = self._name_and_context()
        return getattr(context, name)

    def set(self, value):
        if self.field.readonly:
            raise TypeError(
                "Can't set values on read-only fields "
                "(name=%s, class=%s.%s)"
                % (
                    self.field.__name__,
                    self.context.__class__.__module__,
                    self.context.__class__.__name__,
                )
            )
        (name, context) = self._name_and_context()
        setattr(context, name, value)

    def canAccess(self):
        (name, context) = self._name_and_context()
        if isinstance(context, Proxy):
            return canAccess(context, name)
        return True

    def canWrite(self):
        (name, context) = self._name_and_context()
        if isinstance(context, Proxy):
            return canWrite(context, name)
        return True



from euphorie.client import MessageFactory as _
from euphorie.client.browser.publish import EnableCustomRisks
from euphorie.client.country import IClientCountry
from euphorie.client.sector import IClientSector
from plone import api

import datetime
import logging


log = logging.getLogger(__name__)


def update_custom_risks_module_texts(context):
    """"""
    if not api.portal.get_registry_record("euphorie.allow_user_defined_risks"):
        log.warning(
            "Custom risks are not enabled. Set 'allow_user_defined_risks' to "
            "true in euphorie.ini for enabling them."
        )
        return
    portal = api.portal.get()
    client = portal.client
    count = 0
    for country in client.objectValues():
        if IClientCountry.providedBy(country):
            for sector in country.objectValues():
                if IClientSector.providedBy(sector):
                    for survey in sector.objectValues():
                        is_new = EnableCustomRisks(survey)
                        count += 1
                        custom = getattr(survey, "custom-risks", None)
                        if custom:
                            custom.question = _(
                                "question_other_risks",
                                default=(
                                    "<p><strong>Important:</strong> In "
                                    "order to avoid duplicating risks, "
                                    "we strongly recommend you "
                                    "to go first through all the previous modules, "
                                    "if you have not done it yet.</p>"
                                    "<p>If you don't need to add risks, "
                                    "please continue.</p>"
                                ),
                            )
                        if is_new:
                            survey.published = (
                                survey.id,
                                survey.title,
                                datetime.datetime.now(),
                            )



"""
Model
-----

Mainly: the connection between the ZODB-based content of the backend and the
SQL-based individual session content of the client users.
Also: PAS-based user account for users of the client
"""

from AccessControl.interfaces import IUser
from AccessControl.PermissionRole import _what_not_even_god_should_do
from AccessControl.SecurityInfo import ClassSecurityInfo
from Acquisition import aq_chain
from Acquisition import aq_inner
from Acquisition import aq_parent
from collections import defaultdict
from euphorie.client.client import IClient
from euphorie.client.config import LOCKING_ACTIONS
from euphorie.client.config import LOCKING_SET_ACTIONS
from euphorie.client.enum import Enum
from euphorie.content.protect import safeSQLWrite
from OFS.interfaces import IApplication
from plone import api
from plone.app.event.base import localized_now
from plone.base.utils import safe_text
from plone.memoize import ram
from plone.memoize.instance import memoize
from Products.Five import BrowserView
from sqlalchemy import func
from sqlalchemy import orm
from sqlalchemy import schema
from sqlalchemy import sql
from sqlalchemy import types
from sqlalchemy.event import listen
from sqlalchemy.ext.declarative.extensions import instrument_declarative
from sqlalchemy.orm.decl_base import _declarative_constructor
from sqlalchemy.sql import functions
from z3c.saconfig import Session
from zope.component.hooks import getSite
from zope.deprecation import deprecate
from zope.interface import implementer
from zope.interface import Interface
from zope.sqlalchemy import datamanager

import Acquisition
import bcrypt
import datetime
import logging
import OFS.Traversable
import pytz
import random
import re


BCRYPTED_PATTERN = re.compile(r"^\$2[aby]?\$\d{1,2}\$[.\/A-Za-z0-9]{53}$")

metadata = schema.MetaData()

log = logging.getLogger(__name__)


def _forever_cache_key(func, self, *args):
    """Cache this function call forever."""
    return (func.__name__, self.password, args)


def GenerateSecret(length=32):
    """Return random data."""
    secret = ""
    for i in range(length):
        secret += chr(random.getrandbits(8))
    return secret


class BaseObject(OFS.Traversable.Traversable, Acquisition.Implicit):
    """Zope 2-style base class for our models.

    This base class allows SQL based objects to act like normal Zope 2
    objects. In particular it allows acquisition to find skin objects
    and keeps absolute_url() and getPhysicalPath() working.
    """

    __init__ = _declarative_constructor
    __allow_access_to_unprotected_subobjects__ = True
    __new__ = object.__new__

    security = ClassSecurityInfo()

    def getId(self):
        return str(self.id)

    @security.public
    def getPhysicalPath(self):
        # Get the physical path of the object.
        #
        # We need to override this because the new Zope implementations
        # uses self.id instead of self.getId, which make a big difference in our case
        id = self.getId()

        path = (id,)
        p = aq_parent(aq_inner(self))
        if p is None:
            return path

        func = self.getPhysicalPath.__func__
        while p is not None:
            if func is p.getPhysicalPath.__func__:
                pid = p.getId()
                path = (pid,) + path
                p = aq_parent(aq_inner(p))
            else:
                if IApplication.providedBy(p):
                    path = ("",) + path
                else:
                    path = p.getPhysicalPath() + path
                break

        return path


class SurveyTreeItem(BaseObject):
    """A tree of questions.

    The data is stored in the form of a materialized tree. The path is
    built using a list of item numbers. Each item number has three
    digits and uses 0-prefixing to make sure we can use simple string
    sorting to produce a sorted tree.
    """

    __tablename__ = "tree"
    __table_args__ = (
        schema.UniqueConstraint("session_id", "path"),
        schema.UniqueConstraint("session_id", "zodb_path", "profile_index"),
        {},
    )

    id = schema.Column(types.Integer(), primary_key=True, autoincrement=True)
    session_id = schema.Column(
        types.Integer(),
        schema.ForeignKey("session.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    parent_id = schema.Column(
        types.Integer(),
        schema.ForeignKey("tree.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=True,
        index=True,
    )
    type = schema.Column(
        Enum(["risk", "module"]),
        nullable=False,
        index=True,
    )
    path = schema.Column(
        types.String(40),
        nullable=False,
        index=True,
    )
    has_description = schema.Column(
        types.Boolean(),
        default=False,
        index=True,
    )
    zodb_path = schema.Column(
        types.String(512),
        nullable=False,
    )
    profile_index = schema.Column(
        types.Integer(),
        default=0,
        nullable=False,
    )
    depth = schema.Column(
        types.Integer(),
        default=0,
        nullable=False,
        index=True,
    )
    title = schema.Column(types.Unicode(512))
    postponed = schema.Column(types.Boolean())
    skip_children = schema.Column(
        types.Boolean(),
        default=False,
        nullable=False,
    )

    __mapper_args__ = dict(polymorphic_on=type)

    session = orm.relationship(
        "SurveySession",
        cascade="all",
    )
    #    parent = orm.relationship("SurveyTreeItem", uselist=False)

    @property
    def parent(self):
        # XXX Evil! Figure out why the parent relation does not work
        return self.parent_id and Session.query(SurveyTreeItem).get(self.parent_id)

    def getId(self):
        return self.path[-3:].lstrip("0")

    @property
    def short_path(self):
        def slice(path):
            while path:
                yield path[:3].lstrip("0")
                path = path[3:]

        return slice(self.path)

    @property
    def number(self):
        return ".".join(self.short_path)

    def children(self, filter=None):
        query = (
            Session.query(SurveyTreeItem)
            .filter(SurveyTreeItem.session_id == self.session_id)
            .filter(SurveyTreeItem.depth == self.depth + 1)
        )
        if self.path:
            query = query.filter(SurveyTreeItem.path.like(self.path + "%"))
        if filter is not None:
            query = query.filter(filter)
        return query.order_by(SurveyTreeItem.path)

    def siblings(self, klass=None, filter=None):
        if not self.path:
            return []
        if klass is None:
            klass = SurveyTreeItem
        query = (
            Session.query(klass)
            .filter(klass.session_id == self.session_id)
            .filter(klass.parent_id == self.parent_id)
        )
        if filter is not None:
            query = query.filter(sql.or_(klass.id == self.id, filter))
        return query.order_by(klass.path)

    def addChild(self, item):
        sqlsession = Session()
        query = (
            sqlsession.query(SurveyTreeItem.path)
            .filter(SurveyTreeItem.session_id == self.session_id)
            .filter(SurveyTreeItem.depth == self.depth + 1)
        )
        if self.path:
            query = query.filter(SurveyTreeItem.path.like(self.path + "%"))

        last = query.order_by(SurveyTreeItem.path.desc()).first()
        if not last:
            index = 1
        else:
            index = int(last[0][-3:]) + 1

        item.session = self.session
        item.depth = self.depth + 1
        item.path = (self.path or "") + "%03d" % index
        item.parent_id = self.id
        if self.profile_index != -1:
            item.profile_index = self.profile_index
        sqlsession.add(item)
        self.session.touch()
        return item

    def removeChildren(self, excluded=[]):
        if self.id not in excluded:
            excluded.append(self.id)
        session = Session()
        if self.path:
            filter = sql.and_(
                SurveyTreeItem.session_id == self.session_id,
                SurveyTreeItem.path.like(self.path + "%"),
                sql.not_(SurveyTreeItem.id.in_(excluded)),
            )
        else:
            filter = sql.and_(
                SurveyTreeItem.session_id == self.session_id,
                sql.not_(SurveyTreeItem.id.in_(excluded)),
            )
        removed = session.query(SurveyTreeItem).filter(filter).all()
        session.execute(SurveyTreeItem.__table__.delete().where(filter))
        self.session.touch()
        datamanager.mark_changed(session)
        return removed


class Group(BaseObject):
    __tablename__ = "group"

    group_id = schema.Column(
        types.Unicode(32),
        primary_key=True,
    )
    parent_id = schema.Column(
        types.Unicode(32),
        schema.ForeignKey("group.group_id"),
    )
    short_name = schema.Column(
        types.Unicode(32),
    )
    long_name = schema.Column(
        types.Unicode(256),
    )
    responsible_id = schema.Column(
        types.Unicode(32),
    )
    responsible_fullname = schema.Column(
        types.Unicode(32),
    )

    deactivated = schema.Column(
        types.DateTime,
        nullable=True,
        default=None,
    )

    parent = orm.relationship(
        "Group",
        back_populates="children",
        remote_side=[group_id],
    )

    children = orm.relationship(
        "Group",
        back_populates="parent",
        remote_side=[parent_id],
    )

    accounts = orm.relationship(
        "Account",
        back_populates="group",
    )

    brand = schema.Column(types.String(64))

    sessions = orm.relationship(
        "SurveySession",
        back_populates="group",
        order_by="SurveySession.modified",
        cascade="all, delete-orphan",
    )

    # Allow this class to be subclassed in other projects
    __mapper_args__ = {
        "polymorphic_identity": "euphorie",
        "polymorphic_on": brand,
        "with_polymorphic": "*",
    }

    @property
    def fullname(self):
        """This is the name that will be display in the selectors and in the
        tree widget."""
        title = "{obs}{name}".format(
            obs="[obs.] " if not self.deactivated else "",
            name=self.short_name or self.group_id,
        )
        if self.responsible_fullname:
            title += f", {self.responsible_fullname}"
        return title

    @property
    def descendantids(self):
        """Return all the groups in the hierarchy flattened."""
        structure = self._group_structure
        ids = []

        def get_ids(groupid):
            new_ids = structure[groupid]
            if not new_ids:
                return
            ids.extend(new_ids)
            tuple(map(get_ids, new_ids))

        get_ids(self.group_id)
        return ids

    @property
    def descendants(self):
        """Return all the groups in the hierarchy flattened."""
        return list(
            Session.query(self.__class__).filter(
                self.__class__.group_id.in_(self.descendantids)
            )
        )

    @property
    def parents(self):
        """Return all the groups in the hierarchy flattened."""
        group = self
        parents = []
        while True:
            parent = group.parent
            if not parent:
                return parents
            parents.append(parent)
            group = parent

    @property
    @memoize
    def _group_structure(self):
        """Return a dict like structure with the group ids as keys and the
        children group ids as values."""
        tree = defaultdict(set)
        for groupid, parentid in Session.query(Group.group_id, Group.parent_id).filter(
            Group.parent_id != None  # noqa: E711
        ):
            tree[parentid].add(groupid)
        return tree

    @property
    def acquired_sessions(self):
        """All the session relative to this group and its children."""
        group_ids = [self.group_id]
        group_ids.extend(g.group_id for g in self.descendants)
        return (
            Session.query(SurveySession)
            .filter(SurveySession.group_id.in_(group_ids))
            .all()
        )


class Consultancy(BaseObject):
    """Information about consultancy on a session."""

    __tablename__ = "consultancy"

    session_id = schema.Column(
        schema.ForeignKey("session.id", onupdate="CASCADE", ondelete="CASCADE"),
        primary_key=True,
    )
    account_id = schema.Column(
        schema.ForeignKey("account.id", onupdate="CASCADE", ondelete="SET NULL"),
        nullable=True,
    )

    session = orm.relationship(
        "SurveySession",
        uselist=False,
        back_populates="consultancy",
    )
    account = orm.relationship(
        "Account",
        uselist=False,
        back_populates="consultancy",
    )

    status = schema.Column(
        types.Unicode(255),
        default="pending",
    )


@implementer(IUser)
class Account(BaseObject):
    """A user account.

    Users have to register with euphorie before they can start a survey
    session. A single account can have multiple survey sessions.
    """

    __tablename__ = "account"

    id = schema.Column(
        types.Integer(),
        primary_key=True,
        autoincrement=True,
    )
    loginname = schema.Column(
        types.String(255),
        nullable=False,
        index=True,
        unique=True,
    )
    password = schema.Column(types.Unicode(64))
    tc_approved = schema.Column(types.Integer())
    account_type = schema.Column(
        Enum(["guest", "converted", "full"]),
        default="full",
        nullable=True,
    )
    group_id = schema.Column(
        types.Unicode(32),
        schema.ForeignKey("group.group_id"),
    )

    created = schema.Column(
        types.DateTime,
        nullable=True,
        default=functions.now(),
    )

    last_login = schema.Column(
        types.DateTime,
        nullable=True,
        default=None,
    )

    first_name = schema.Column(
        types.Unicode(),
        nullable=True,
        default=None,
    )
    last_name = schema.Column(
        types.Unicode(),
        nullable=True,
        default=None,
    )
    consultancy = orm.relationship(
        "Consultancy",
        uselist=False,
        back_populates="account",
    )

    group = orm.relationship(
        Group,
        back_populates="accounts",
    )

    @property
    def groups(self):
        group = self.group
        if not group:
            return []
        groups = [group]
        groups.extend(group.descendants)
        return groups

    @property
    def acquired_sessions(self):
        """The session the account acquires because he belongs to a group."""
        group = self.group
        if not group:
            return []
        return list(group.acquired_sessions)

    @property
    def group_sessions(self):
        """The session the account acquires because he belongs to a group."""
        group = self.group
        if not group:
            return []
        return list(group.sessions)

    @property
    def email(self):
        """Email addresses are used for login, return the login."""
        return self.loginname

    @property
    def login(self):
        """This synchs naming with :obj:`euphorie.content.user.IUser` and is
        needed by the authentication tools."""
        return self.loginname

    @property
    def title(self):
        """Return the joined first_name and last_name of the account if
        present.

        If they are not fallback to the loginname
        """
        return (
            " ".join((self.first_name or "", self.last_name or "")).strip()
            or self.loginname
        )

    def getUserName(self):
        """Return the login name."""
        return self.loginname

    def getGroups(self):
        return ["AuthenticatedUsers"]

    def getRoles(self):
        """Return all global roles for this user."""

        return ("EuphorieUser",)

    def getRolesInContext(self, object):
        """Return the roles of the user in the current context (same as
        :obj:`getRoles`).
        """
        return self.getRoles()

    def getDomains(self):
        return []

    def addPropertysheet(self, propfinder_id, data):
        pass

    def _addGroups(self, group_ids):
        pass

    def _addRoles(self, role_ids):
        pass

    def has_permission(self, perm, context):
        """Check if the user has a permission in a context."""
        return perm == "Euphorie: View a Survey"

    def allowed(self, context, object_roles=None):
        """Check if this account has any of the requested roles in the context
        of `object`."""
        if object_roles is _what_not_even_god_should_do:
            return False

        if object_roles is None:
            return True

        for obj in aq_chain(aq_inner(context)):
            if IClient.providedBy(obj):
                allowed_roles = {"Anonymous", "Authenticated", "EuphorieUser", "Reader"}
                return bool(allowed_roles & set(object_roles))

        return False

    @ram.cache(_forever_cache_key)
    def verify_password(self, password):
        """Verify the given password against the one stored in the account
        table."""
        if not password:
            return False
        if not isinstance(password, str):
            return False
        if password == self.password:
            return True
        password = safe_text(password)
        return bcrypt.checkpw(password, self.password)

    def hash_password(self):
        """Hash the account password using bcrypt."""
        try:
            password = self.password
        except AttributeError:
            return
        if not password:
            return
        password = safe_text(password)
        if BCRYPTED_PATTERN.match(password):
            # The password is already encrypted, do not encrypt it again
            # XXX this is broken with passwords that are actually an hash
            return
        self.password = safe_text(
            bcrypt.hashpw(
                password,
                bcrypt.gensalt(),
            )
        )


def account_before_insert_subscriber(mapper, connection, account):
    account.hash_password()


account_before_update_subscriber = account_before_insert_subscriber

listen(Account, "before_insert", account_before_insert_subscriber)
listen(Account, "before_update", account_before_update_subscriber)


class AccountChangeRequest(BaseObject):
    __tablename__ = "account_change"

    id = schema.Column(types.String(16), primary_key=True, nullable=False)
    account_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Account.id, onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
        unique=True,
    )
    account = orm.relationship(
        Account,
        back_populates="change_request",
    )
    value = schema.Column(
        types.String(255),
        nullable=False,
    )
    expires = schema.Column(
        types.DateTime(),
        nullable=False,
    )


Account.change_request = orm.relationship(
    AccountChangeRequest,
    back_populates="account",
    cascade="all, delete-orphan",
    uselist=False,
)


class ISurveySession(Interface):
    """Marker interface for a SurveySession object."""


@implementer(ISurveySession)
class SurveySession(BaseObject):
    """Information about a user's session."""

    __tablename__ = "session"

    id = schema.Column(types.Integer(), primary_key=True, autoincrement=True)
    brand = schema.Column(types.String(64))
    account_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Account.id, onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    last_modifier_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Account.id, onupdate="CASCADE", ondelete="CASCADE"),
        nullable=True,
        index=False,
    )
    last_publisher_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Account.id, onupdate="CASCADE", ondelete="CASCADE"),
        nullable=True,
        index=False,
    )
    group_id = schema.Column(
        types.Unicode(32),
        schema.ForeignKey("group.group_id"),
    )
    title = schema.Column(types.Unicode(512))
    created = schema.Column(
        types.DateTime,
        nullable=False,
        default=functions.now(),
    )
    modified = schema.Column(
        types.DateTime,
        nullable=False,
        default=functions.now(),
    )
    refreshed = schema.Column(
        types.DateTime,
        nullable=False,
        default=functions.now(),
    )

    published = schema.Column(
        types.DateTime,
        nullable=True,
        default=None,
    )

    archived = schema.Column(
        types.DateTime(timezone=True),
        nullable=True,
        default=None,
    )

    zodb_path = schema.Column(types.String(512), nullable=False)

    report_comment = schema.Column(types.UnicodeText())

    account = orm.relationship(
        Account,
        back_populates="sessions",
        foreign_keys=[account_id],
    )
    last_modifier = orm.relationship(
        Account,
        foreign_keys=[last_modifier_id],
    )
    last_publisher = orm.relationship(
        Account,
        foreign_keys=[last_publisher_id],
    )

    group = orm.relationship(
        Group,
        back_populates="sessions",
    )

    consultancy = orm.relationship(
        "Consultancy",
        uselist=False,
        back_populates="session",
    )

    migrated = schema.Column(
        types.DateTime,
        nullable=False,
        default=functions.now(),
    )

    # Allow this class to be subclassed in other projects
    __mapper_args__ = {
        "polymorphic_identity": "euphorie",
        "polymorphic_on": brand,
        "with_polymorphic": "*",
    }

    @property
    def is_archived(self):
        archived = self.archived
        if not archived:
            return False
        return archived <= localized_now()

    @property
    def last_locking_event(self):
        """Return the last event relative to locking"""
        query = (
            Session.query(SessionEvent)
            .filter(
                SessionEvent.action.in_(LOCKING_ACTIONS),
                SessionEvent.session_id == self.id,
            )
            .order_by(SessionEvent.time.desc())
        )
        return query.first()

    @property
    def last_validation_event(self):
        """Return the last event relative to validation"""
        query = (
            Session.query(SessionEvent)
            .filter(
                SessionEvent.action.in_(
                    (
                        "validation_requested",
                        "validated",
                        "invalidated",
                    )
                ),
                SessionEvent.session_id == self.id,
            )
            .order_by(SessionEvent.time.desc())
        )
        return query.first()

    @property
    def is_validated(self):
        """Check if the session is validated."""
        event = self.last_validation_event
        if not event:
            return False
        return event.action == "validated"

    @property
    def is_locked(self):
        """Check if the session is locked."""
        if self.is_validated:
            return True

        event = self.last_locking_event
        if not event:
            return False
        return event.action in LOCKING_SET_ACTIONS

    @property
    @deprecate(
        "Deprecated in version 15.0.0.dev0. "
        "You might want to use self.is_locked instead."
    )
    def review_state(self):
        """Check if it the published column.

        If it has return 'published' otherwise return 'private'
        """
        return "published" if self.is_locked else "private"

    def hasTree(self):
        return bool(
            Session.query(SurveyTreeItem).filter(SurveyTreeItem.session == self).count()
        )

    def reset(self):
        Session.query(SurveyTreeItem).filter(SurveyTreeItem.session == self).delete()
        self.created = self.modified = datetime.datetime.now()

    def touch(self):
        self.last_modifier = get_current_account()
        self.modified = datetime.datetime.now()

    def refresh_survey(self, survey=None):
        """Mark the session with the current date to indicate that is has been
        refreshed with the latest version of the Survey (from Zope).

        If survey is passed, update all titles in the tree, based on the
        CMS version of the survey, i.e. update all titles of modules and
        risks. Those are used in the navigation. If a title change is
        the only change in the CMS, the survey session is not re-
        created. Therefore this method ensures that the titles are
        updated where necessary.
        """
        if survey:
            query = Session.query(SurveyTreeItem).filter(
                SurveyTreeItem.session_id == self.id
            )
            tree = query.all()
            for item in tree:
                if item.zodb_path.find("custom-risks") >= 0:
                    continue
                zodb_item = survey.restrictedTraverse(item.zodb_path.split("/"), None)
                # Don't update session-specific instances of profilequestions
                if (
                    zodb_item.portal_type == "euphorie.profilequestion"
                    and item.profile_index > -1
                ):
                    continue
                if zodb_item and zodb_item.title != item.title:
                    item.title = zodb_item.title
                    safeSQLWrite(item)
        self.refreshed = datetime.datetime.now()
        safeSQLWrite(self)

    def update_measure_types(self, survey):
        """Update measure types in the session according to changes in the
        tool.

        Specifically, if an `in_place_standard` measure is deleted in
        the tool, it disappears from the identification phase of the
        session unless we change its type to `in_place_custom`.
        """
        in_place_standard_measures = (
            Session.query(Risk, ActionPlan)
            .filter(Risk.id == ActionPlan.risk_id)
            .filter(Risk.session_id == self.id)
            .filter(ActionPlan.plan_type == "in_place_standard")
            .all()
        )
        for risk, measure in in_place_standard_measures:
            risk_zodb = survey.restrictedTraverse(risk.zodb_path.split("/"))
            solution_ids_zodb = [sol.id for sol in risk_zodb._solutions]
            if measure.solution_id not in solution_ids_zodb:
                # The measure is in the session but not in the tool. It has probably
                # been deleted. Keep it visible by making it a custom measure.
                measure.plan_type = "in_place_custom"

    def addChild(self, item):
        sqlsession = Session()
        query = (
            sqlsession.query(SurveyTreeItem.path)
            .filter(SurveyTreeItem.session_id == self.id)
            .filter(SurveyTreeItem.depth == 1)
            .order_by(SurveyTreeItem.path.desc())
        )
        last = query.first()
        if not last:
            index = 1
        else:
            index = int(last[0][-3:]) + 1

        item.session = self
        item.depth = 1
        item.path = "%03d" % index
        item.parent_id = None
        sqlsession.add(item)
        self.touch()
        return item

    def children(self, filter=None):
        query = (
            Session.query(SurveyTreeItem)
            .filter(SurveyTreeItem.session_id == self.id)
            .filter(SurveyTreeItem.depth == 1)
        )
        if filter is not None:
            query = query.filter(filter)
        return query.order_by(SurveyTreeItem.path)

    def copySessionData(self, other):
        """Copy all user data from another session to this one."""
        session = Session()

        # Copy all tree data to the new session (skip_children and postponed)
        old_tree = orm.aliased(SurveyTreeItem, name="old_tree")
        in_old_tree = sql.and_(
            old_tree.session_id == other.id,
            SurveyTreeItem.zodb_path == old_tree.zodb_path,
            SurveyTreeItem.profile_index == old_tree.profile_index,
        )
        skip_children = sql.select([old_tree.skip_children], in_old_tree).limit(1)
        postponed = sql.select([old_tree.postponed], in_old_tree).limit(1)
        new_items = (
            session.query(SurveyTreeItem)
            .filter(SurveyTreeItem.session == self)
            .filter(sql.exists(sql.select([old_tree.id]).where(in_old_tree)))
        )
        new_items.update(
            {"skip_children": skip_children, "postponed": postponed},
            synchronize_session=False,
        )

        # Mandatory modules must have skip_children=False. It's possible that
        # the module was optional with skip_children=True and now after the
        # update it's mandatory. So we must check and correct.
        # In case a risk was marked as "always present", be sure its
        # identification gets set to 'no'
        preset_to_no = []
        survey = getSite()["client"].restrictedTraverse(self.zodb_path)
        for item in new_items.all():
            if item.type == "risk":
                if item.identification == "no":
                    preset_to_no.append(item.risk_id)

            elif item.type == "module":
                module = survey.restrictedTraverse(item.zodb_path.split("/"))
                if not module.optional:
                    item.skip_children = False

        # Copy all risk data to the new session
        # This triggers a "Only update via a single table query is currently
        # supported" error with SQLAlchemy 0.6.6
        # old_risk = orm.aliased(Risk.__table__, name='old_risk')
        # is_old_risk = sql.and_(in_old_tree, old_tree.id == old_risk.id)
        # identification = sql.select([old_risk.identification], is_old_risk)
        # new_risks = session.query(Risk)\
        #         .filter(Risk.session == self)\
        #         .filter(sql.exists(
        #             sql.select([SurveyTreeItem.id]).where(sql.and_(
        #                     SurveyTreeItem.id == Risk.id,
        #                     sql.exists([old_tree.id]).where(sql.and_(
        #                         in_old_tree, old_tree.type == 'risk'))))))
        # new_risks.update({'identification': identification},
        #         synchronize_session=False)

        skip_preset_to_no_clause = ""
        if len(preset_to_no):
            skip_preset_to_no_clause = "old_risk.risk_id not in %s AND" % (
                str([str(x) for x in preset_to_no]).replace("[", "(").replace("]", ")")
            )
        statement = """\
        UPDATE RISK
        SET identification = old_risk.identification,
            frequency = old_risk.frequency,
            effect = old_risk.effect,
            probability = old_risk.probability,
            priority = old_risk.priority,
            existing_measures = old_risk.existing_measures,
            comment = old_risk.comment,
            scaled_answer = old_risk.scaled_answer
        FROM risk AS old_risk JOIN tree AS old_tree ON old_tree.id=old_risk.id, tree
        WHERE tree.id=risk.id AND
              %(skip_preset_to_no_clause)s
              tree.session_id=%(new_sessionid)s AND
              old_tree.session_id=%(old_sessionid)s AND
              old_tree.zodb_path=tree.zodb_path AND
              old_tree.profile_index=tree.profile_index;
        """ % dict(  # noqa: E501
            old_sessionid=other.id,
            new_sessionid=self.id,
            skip_preset_to_no_clause=skip_preset_to_no_clause,
        )
        session.execute(statement)

        statement = """\
        INSERT INTO action_plan (risk_id, action_plan, prevention_plan, action,
                                        requirements, responsible, budget, plan_type,
                                        planning_start, planning_end,
                                        solution_id, used_in_training)
               SELECT new_tree.id,
                      action_plan.action_plan,
                      action_plan.prevention_plan,
                      action_plan.action,
                      action_plan.requirements,
                      action_plan.responsible,
                      action_plan.budget,
                      action_plan.plan_type,
                      action_plan.planning_start,
                      action_plan.planning_end,
                      action_plan.solution_id,
                      action_plan.used_in_training
               FROM action_plan JOIN risk ON action_plan.risk_id=risk.id
                                JOIN tree ON tree.id=risk.id,
                    tree AS new_tree
               WHERE tree.session_id=%(old_sessionid)d AND
                     new_tree.session_id=%(new_sessionid)d AND
                     tree.zodb_path=new_tree.zodb_path AND
                     tree.profile_index=new_tree.profile_index;
            """ % {
            "old_sessionid": other.id,
            "new_sessionid": self.id,
        }
        session.execute(statement)

        # Copy over previous session metadata. Specifically, we don't want to
        # create a new modification timestamp, just because the underlying
        # survey was updated.
        statement = """\
        UPDATE session
        SET
            modified = old_session.modified,
            created = old_session.created,
            last_modifier_id = old_session.last_modifier_id
        FROM session as old_session
        WHERE
            old_session.id=%(old_sessionid)d AND
            session.id=%(new_sessionid)d
        """ % {
            "old_sessionid": other.id,
            "new_sessionid": self.id,
        }
        session.execute(statement)

        session.query(Company).filter(Company.session == other).update(
            {"session_id": self.id}, synchronize_session=False
        )

    @classmethod
    def get_account_filter(cls, account=None):
        """Filter only the sessions for the given account.

        :param acount: True means current account.
            A falsish value means do not filter.
            Otherwise try to interpret the user input:
            a string or an int means the account_id should be that value,
            an object account will be used to extract the account id,
            from an iterable we will try to extract the account ids
        """
        # TODO: this is too complex
        include_organisation_members = False
        if account is True:
            include_organisation_members = True
            account = get_current_account()

        if isinstance(account, Account):
            account = account.id

        if not account:
            return False

        if not include_organisation_members and isinstance(account, (int, (str,))):
            return cls.account_id == account

        if include_organisation_members:
            account_ids = {account}
            # Add the owner id of the organisations where the account is member of
            owner_memberships = Session.query(OrganisationMembership.owner_id).filter(
                OrganisationMembership.member_id == account
            )
            account_ids.update(membership.owner_id for membership in owner_memberships)
        else:
            try:
                # This works when we pass an iterable of accounts or ids
                account_ids = {getattr(item, "id", item) for item in account}
            except TypeError:
                # this happens when account is not an iterable
                log.error("Cannot understand the account parameter: %r", account)
                raise
            account_ids = {
                item for item in account_ids if item and isinstance(item, (int, (str,)))
            }

        if not account_ids:
            return False

        if len(account_ids) == 1:
            for account_id in account_ids:
                return cls.get_account_filter(account_id)

        return cls.account_id.in_(account_ids)

    @classmethod
    def get_group_filter(cls, group=None):
        """Filter only the sessions for the given group.

        :param group: True means the current account's group.
            A falsish value means do not filter.
            Otherwise try to interpret the user input:
            a string or an int means the group_id should be that value,
            an object group will be used to extract the group id,
            and from an iterable we will try to extract the group ids
        """
        if group is True:
            group = getattr(get_current_account(), "group_id", None)

        if isinstance(group, Group):
            group = group.group_id

        if not group:
            return False

        if isinstance(group, (int, (str,))):
            return cls.group_id == group

        try:
            group_ids = {getattr(item, "group_id", item) for item in group}
        except TypeError:
            log.error("Cannot understand the group parameter: %r", group)
            raise

        group_ids = {
            item for item in group_ids if item and isinstance(item, (int, (str,)))
        }
        if not group_ids:
            return False

        if len(group_ids) == 1:
            for group_id in group_ids:
                return cls.get_group_filter(group_id)

        return cls.group_id.in_(group_ids)

    @classmethod
    def get_archived_filter(cls):
        """Filter out sessions that are archived."""
        return sql.or_(
            cls.archived >= localized_now(), cls.archived == None  # noqa: E711
        )

    @classmethod
    def _get_context_tools(cls, context):
        """Return the set of tools we can find under this context."""
        if not context:
            return set()

        # Check the path relative to the client folder
        if context.portal_type == "Plone Site":
            context = context.client

        if context.portal_type == "euphorie.survey":
            return {context}

        portal_type_filter = {
            "portal_type": [
                "euphorie.clientcountry",
                "euphorie.clientsector",
                "euphorie.survey",
            ]
        }

        surveys = set()

        def _add_survey(container):
            for obj in container.listFolderContents(portal_type_filter):
                if obj.portal_type == "euphorie.survey":
                    surveys.add(obj)
                else:
                    _add_survey(obj)

        _add_survey(context)
        return surveys

    @classmethod
    def get_context_filter(cls, context):
        """Filter sessions under this context using the zodb_path column."""
        surveys = cls._get_context_tools(context)
        if not surveys:
            return False

        return cls.zodb_path.in_(
            {safe_text("/".join(survey.getPhysicalPath()[-3:])) for survey in surveys}
        )

    @property
    def tool(self):
        client = api.portal.get().client
        return client.restrictedTraverse(str(self.zodb_path), None)

    @property
    def traversed_session(self):
        return self.tool.restrictedTraverse("++session++%s" % self.id)

    def absolute_url(self):
        """The URL for this session is based on the tool's URL.

        To it (if it can be fetched) we add a traverser with the session
        id.
        """
        client = api.portal.get().client
        tool = client.unrestrictedTraverse(self.zodb_path, None)
        if tool is None:
            raise ValueError("No tool found for session %s" % self.id)
        return f"{tool.absolute_url()}/++session++{self.id}"

    @property
    def country(self):
        return str(self.zodb_path).split("/")[0]

    @property
    def completion_percentage(self):
        module_query = (
            Session.query(SurveyTreeItem)
            .filter(SurveyTreeItem.session_id == self.id)
            .filter(SurveyTreeItem.type == "module")
        ).order_by(SurveyTreeItem.path)

        good_module_ids = set()
        bad_module_ids = set()
        for module in module_query:
            if module.parent_id in bad_module_ids or module.skip_children:
                bad_module_ids.add(module.id)
            else:
                good_module_ids.add(module.id)

        if not good_module_ids:
            return 0

        total_risks_query = Session.query(Risk).filter(
            Risk.parent_id.in_(good_module_ids)
        )
        total = total_risks_query.count()

        if not total:
            return 0

        answered = float(
            total_risks_query.filter(Risk.identification != None).count()  # noqa: E711
        )

        completion_percentage = int(round(answered / total * 100.0))
        return completion_percentage


Account.sessions = orm.relationship(
    SurveySession,
    back_populates="account",
    foreign_keys=[SurveySession.account_id],
    cascade="all, delete-orphan",
)


class SessionEvent(BaseObject):
    """Data table to record events happening on sessions."""

    __tablename__ = "session_event"

    id = schema.Column(types.Integer(), primary_key=True, autoincrement=True)
    time = schema.Column(types.DateTime(), nullable=False, default=func.now())
    account_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Account.id, onupdate="CASCADE"),
        nullable=True,
    )
    account = orm.relationship(Account)
    session_id = schema.Column(
        types.Integer(),
        schema.ForeignKey("session.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )
    session = orm.relationship(SurveySession, back_populates="events")
    action = schema.Column(types.Unicode(32))
    note = schema.Column(types.Unicode)


Account.session_events = orm.relationship(SessionEvent, back_populates="account")


SurveySession.events = orm.relationship(
    SessionEvent, back_populates="session", cascade="all,delete-orphan"
)


class SessionRedirect(BaseObject):
    """Mapping of old deleted sessions to their new rebuilt counterparts"""

    __tablename__ = "session_redirect"

    old_session_id = schema.Column(types.Integer(), primary_key=True, nullable=False)
    new_session_id = schema.Column(types.Integer(), nullable=False)


class Company(BaseObject):
    """Information about a company."""

    __tablename__ = "company"

    id = schema.Column(types.Integer(), primary_key=True, autoincrement=True)
    session_id = schema.Column(
        types.Integer(),
        schema.ForeignKey("session.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    session = orm.relationship(SurveySession, back_populates="company")

    country = schema.Column(types.String(3))
    employees = schema.Column(Enum([None, "1-9", "10-49", "50-249", "250+"]))
    conductor = schema.Column(Enum([None, "staff", "third-party", "both"]))
    referer = schema.Column(
        Enum(
            [
                None,
                "employers-organisation",
                "trade-union",
                "national-public-institution",
                "eu-institution",
                "health-safety-experts",
                "other",
            ]
        )
    )
    workers_participated = schema.Column(types.Boolean())
    needs_met = schema.Column(types.Boolean())
    recommend_tool = schema.Column(types.Boolean())
    timestamp = schema.Column(types.DateTime(), nullable=True)


SurveySession.company = orm.relationship(
    Company, back_populates="session", cascade="all,delete-orphan", uselist=False
)


class Module(SurveyTreeItem):
    """A module.

    This is a dummy object needed to be able to put modules in the
    survey tree.
    """

    __tablename__ = "module"
    __mapper_args__ = dict(polymorphic_identity="module")

    sql_module_id = schema.Column(
        "id",
        types.Integer(),
        schema.ForeignKey(SurveyTreeItem.id, onupdate="CASCADE", ondelete="CASCADE"),
        primary_key=True,
    )
    module_id = schema.Column(types.String(16), nullable=False)
    solution_direction = schema.Column(types.Boolean(), default=False)


class Risk(SurveyTreeItem):
    """Answer to risk."""

    __tablename__ = "risk"
    __mapper_args__ = dict(polymorphic_identity="risk")

    sql_risk_id = schema.Column(
        "id",
        types.Integer(),
        schema.ForeignKey(SurveyTreeItem.id, onupdate="CASCADE", ondelete="CASCADE"),
        primary_key=True,
    )
    risk_id = schema.Column(types.String(16), nullable=True)
    risk_type = schema.Column(
        Enum(["risk", "policy", "top5"]), default="risk", nullable=False, index=True
    )
    #: Skip-evaluation flag. This is only used to indicate if the sector
    #: set the evaluation method to `fixed`, not for policy behaviour
    #: such as not evaluation top-5 risks. That policy behaviour is
    #: handled via the question_filter on client views so it can be modified
    #: in custom deployments.
    skip_evaluation = schema.Column(types.Boolean(), default=False, nullable=False)
    is_custom_risk = schema.Column(types.Boolean(), default=False, nullable=False)
    identification = schema.Column(Enum([None, "yes", "no", "n/a"]))
    frequency = schema.Column(types.Integer())
    effect = schema.Column(types.Integer())
    probability = schema.Column(types.Integer())
    priority = schema.Column(Enum([None, "low", "medium", "high"]))
    comment = schema.Column(types.UnicodeText())
    existing_measures = schema.Column(types.UnicodeText())
    training_notes = schema.Column(types.UnicodeText())
    custom_description = schema.Column(types.UnicodeText())
    image_data = schema.Column(types.LargeBinary())
    image_data_scaled = schema.Column(types.LargeBinary())
    image_filename = schema.Column(types.UnicodeText())
    scaled_answer = schema.Column(types.UnicodeText())

    @memoize
    def measures_of_type(self, plan_type):
        query = (
            Session.query(ActionPlan)
            .filter(
                sql.and_(ActionPlan.risk_id == self.id),
                ActionPlan.plan_type == plan_type,
            )
            .order_by(ActionPlan.id)
        )
        return query.all()

    @property
    def standard_measures(self):
        return self.measures_of_type("measure_standard")

    @property
    def custom_measures(self):
        return self.measures_of_type("measure_custom")

    @property
    def in_place_standard_measures(self):
        return self.measures_of_type("in_place_standard")

    @property
    def in_place_custom_measures(self):
        return self.measures_of_type("in_place_custom")


class ActionPlan(BaseObject):
    """Action plans for a known risk."""

    __tablename__ = "action_plan"

    id = schema.Column(types.Integer(), primary_key=True, autoincrement=True)
    risk_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Risk.id, onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
        index=True,
    )
    action_plan = schema.Column(types.UnicodeText())
    prevention_plan = schema.Column(types.UnicodeText())
    # The column "action" is the synthesis of "action_plan" and "prevention_plan"
    action = schema.Column(types.UnicodeText())
    requirements = schema.Column(types.UnicodeText())
    responsible = schema.Column(types.Unicode(256))
    budget = schema.Column(types.Integer())
    planning_start = schema.Column(types.Date())
    planning_end = schema.Column(types.Date())
    reference = schema.Column(types.Text())
    plan_type = schema.Column(
        Enum(
            [
                "measure_custom",
                "measure_standard",
                "in_place_standard",
                "in_place_custom",
            ]
        ),
        nullable=False,
        index=True,
        default="measure_custom",
    )
    solution_id = schema.Column(types.Unicode(20))
    used_in_training = schema.Column(
        types.Boolean(),
        default=True,
        index=True,
    )

    risk = orm.relationship(Risk, back_populates="action_plans")


Risk.action_plans = orm.relationship(
    ActionPlan, back_populates="risk", cascade="all, delete-orphan"
)


class Training(BaseObject):
    """Data table to record trainings."""

    __tablename__ = "training"

    id = schema.Column(types.Integer(), primary_key=True, autoincrement=True)
    time = schema.Column(types.DateTime(), nullable=True, default=func.now())
    account_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Account.id, onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )
    account = orm.relationship(Account, back_populates="trainings")
    session_id = schema.Column(
        types.Integer(),
        schema.ForeignKey("session.id", onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )
    session = orm.relationship(
        "SurveySession",
        back_populates="trainings",
    )
    answers = schema.Column(types.Unicode, default="[]")
    status = schema.Column(types.Unicode)


Account.trainings = orm.relationship(
    Training, back_populates="account", cascade="all, delete-orphan"
)


SurveySession.trainings = orm.relationship(
    Training, back_populates="session", cascade="all, delete-orphan"
)


class Organisation(BaseObject):
    """A table to store some data about an organisation."""

    __tablename__ = "organisation"

    organisation_id = schema.Column(
        types.Integer(),
        primary_key=True,
        autoincrement=True,
    )
    owner_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Account.id, onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )
    title = schema.Column(types.UnicodeText())
    image_data = schema.Column(types.LargeBinary())
    image_data_scaled = schema.Column(types.LargeBinary())
    image_filename = schema.Column(types.UnicodeText())
    owner = orm.relationship(Account, back_populates="organisation")


Account.organisation = orm.relationship(
    Organisation, back_populates="owner", cascade="all, delete-orphan", uselist=False
)


class OrganisationMembership(BaseObject):
    """This table wants to mimic the concept of an organisation for Euphorie.

    The goal is to share permissions to work on sessions from another
    user.
    """

    __tablename__ = "organisation_membership"

    organisation_id = schema.Column(
        types.Integer(),
        primary_key=True,
        autoincrement=True,
    )
    owner_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Account.id, onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )
    member_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Account.id, onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )
    member_role = schema.Column(types.UnicodeText())


class NotificationSubscription(BaseObject):
    __tablename__ = "notification_subscription"

    id = schema.Column(types.Integer(), primary_key=True, autoincrement=True)
    account_id = schema.Column(
        types.Integer(),
        schema.ForeignKey(Account.id, onupdate="CASCADE", ondelete="CASCADE"),
        nullable=False,
    )
    category = schema.Column(types.String(512), nullable=False)
    enabled = schema.Column(types.Boolean(), nullable=False, default=False)


_instrumented = False
if not _instrumented:
    metadata._decl_registry = {}
    for cls in [
        Consultancy,
        SurveyTreeItem,
        SurveySession,
        SessionEvent,
        SessionRedirect,
        Module,
        Risk,
        ActionPlan,
        Group,
        Account,
        AccountChangeRequest,
        Company,
        Training,
        Organisation,
        OrganisationMembership,
        NotificationSubscription,
    ]:
        instrument_declarative(cls, metadata._decl_registry, metadata)
    orm.configure_mappers()
    _instrumented = True

schema.Index("tree_session_path", SurveyTreeItem.session_id, SurveyTreeItem.path)
schema.Index(
    "tree_zodb_path",
    SurveyTreeItem.session_id,
    SurveyTreeItem.profile_index,
    SurveyTreeItem.zodb_path,
)


def _SKIPPED_PARENTS_factory():
    # XXX This can be optimized by doing short-circuit on parent.type!=module
    parent = orm.aliased(SurveyTreeItem)
    return sql.exists().where(
        sql.and_(
            parent.session_id == SurveyTreeItem.session_id,
            SurveyTreeItem.depth > parent.depth,
            SurveyTreeItem.path.like(parent.path + "%"),
            parent.skip_children == True,  # noqa: E712
        )
    )


SKIPPED_PARENTS = _SKIPPED_PARENTS_factory()


NO_CUSTOM_RISKS_FILTER = sql.not_(
    sql.and_(
        SurveyTreeItem.type == "risk",
        sql.exists(
            sql.select([Risk.sql_risk_id]).where(
                sql.and_(
                    Risk.sql_risk_id == SurveyTreeItem.id,
                    Risk.is_custom_risk == True,  # noqa: E712
                )
            )
        ),
    )
)

RISK_OR_MODULE_WITH_DESCRIPTION_FILTER = sql.or_(
    SurveyTreeItem.type != "module", SurveyTreeItem.has_description
)


# Used by tno.euphorie
def _MODULE_WITH_RISK_FILTER_factory():
    child_node = orm.aliased(SurveyTreeItem)
    return sql.and_(
        SurveyTreeItem.type == "module",
        SurveyTreeItem.skip_children == False,  # noqa: E712
        sql.exists(
            sql.select([child_node.id]).where(
                sql.and_(
                    child_node.session_id == SurveyTreeItem.session_id,
                    child_node.id == Risk.sql_risk_id,
                    child_node.type == "risk",
                    Risk.identification == "no",
                    child_node.depth > SurveyTreeItem.depth,
                    child_node.path.like(SurveyTreeItem.path + "%"),
                )
            )
        ),
    )


MODULE_WITH_RISK_FILTER = _MODULE_WITH_RISK_FILTER_factory()


def _MODULE_WITH_RISK_OR_TOP5_FILTER_factory():
    child_node = orm.aliased(SurveyTreeItem)
    return sql.and_(
        SurveyTreeItem.type == "module",
        SurveyTreeItem.skip_children == False,  # noqa: E712
        sql.exists(
            sql.select([child_node.id]).where(
                sql.and_(
                    child_node.session_id == SurveyTreeItem.session_id,
                    child_node.id == Risk.sql_risk_id,
                    child_node.type == "risk",
                    sql.or_(Risk.identification == "no", Risk.risk_type == "top5"),
                    child_node.depth > SurveyTreeItem.depth,
                    child_node.path.like(SurveyTreeItem.path + "%"),
                )
            )
        ),
    )


MODULE_WITH_RISK_OR_TOP5_FILTER = _MODULE_WITH_RISK_OR_TOP5_FILTER_factory()


def _MODULE_WITH_RISK_TOP5_TNO_FILTER_factory():
    child_node = orm.aliased(SurveyTreeItem)
    return sql.and_(
        SurveyTreeItem.type == "module",
        SurveyTreeItem.skip_children == False,  # noqa: E712
        sql.exists(
            sql.select([child_node.id]).where(
                sql.and_(
                    child_node.session_id == SurveyTreeItem.session_id,
                    child_node.id == Risk.sql_risk_id,
                    child_node.type == "risk",
                    sql.or_(
                        Risk.identification == "no",
                        sql.and_(
                            Risk.risk_type == "top5",
                            sql.or_(
                                sql.not_(Risk.identification.in_(["n/a", "yes"])),
                                Risk.identification is None,
                            ),
                        ),
                    ),
                    child_node.depth > SurveyTreeItem.depth,
                    child_node.path.like(SurveyTreeItem.path + "%"),
                )
            )
        ),
    )


# Used by tno.euphorie
MODULE_WITH_RISK_TOP5_TNO_FILTER = _MODULE_WITH_RISK_TOP5_TNO_FILTER_factory()


def _MODULE_WITH_RISK_NO_TOP5_NO_POLICY_DO_EVALUTE_FILTER_factory():
    child_node = orm.aliased(SurveyTreeItem)
    return sql.and_(
        SurveyTreeItem.type == "module",
        SurveyTreeItem.skip_children == False,  # noqa: E712
        sql.exists(
            sql.select([child_node.id]).where(
                sql.and_(
                    child_node.session_id == SurveyTreeItem.session_id,
                    child_node.id == Risk.sql_risk_id,
                    child_node.type == "risk",
                    sql.not_(Risk.risk_type.in_(["top5", "policy"])),
                    sql.not_(Risk.skip_evaluation is True),
                    Risk.identification == "no",
                    child_node.depth > SurveyTreeItem.depth,
                    child_node.path.like(SurveyTreeItem.path + "%"),
                )
            )
        ),
    )


MODULE_WITH_RISK_NO_TOP5_NO_POLICY_DO_EVALUTE_FILTER = (
    _MODULE_WITH_RISK_NO_TOP5_NO_POLICY_DO_EVALUTE_FILTER_factory()
)

# Used by tno.euphorie
RISK_PRESENT_FILTER = sql.and_(
    SurveyTreeItem.type == "risk",
    sql.exists(
        sql.select([Risk.sql_risk_id]).where(
            sql.and_(Risk.sql_risk_id == SurveyTreeItem.id, Risk.identification == "no")
        )
    ),
)


def _RISK_PRESENT_FILTER_TOP5_TNO_FILTER_factory():
    Risk_ = orm.aliased(Risk)
    return sql.and_(
        SurveyTreeItem.type == "risk",
        sql.exists(
            sql.select([Risk_.sql_risk_id]).where(
                sql.and_(
                    Risk_.sql_risk_id == SurveyTreeItem.id,
                    sql.or_(
                        Risk_.identification == "no",
                        sql.and_(
                            Risk_.risk_type == "top5",
                            sql.or_(
                                sql.not_(Risk_.identification.in_(["n/a", "yes"])),
                                Risk_.identification == None,  # noqa: E711
                            ),
                        ),
                    ),
                )
            )
        ),
    )


RISK_PRESENT_FILTER_TOP5_TNO_FILTER = _RISK_PRESENT_FILTER_TOP5_TNO_FILTER_factory()


def _RISK_PRESENT_OR_TOP5_FILTER_factory():
    Risk_ = orm.aliased(Risk)
    return sql.and_(
        SurveyTreeItem.type == "risk",
        sql.exists(
            sql.select([Risk_.sql_risk_id]).where(
                sql.and_(
                    Risk_.sql_risk_id == SurveyTreeItem.id,
                    sql.or_(
                        Risk_.identification == "no",
                        Risk_.risk_type == "top5",
                    ),
                )
            )
        ),
    )


RISK_PRESENT_OR_TOP5_FILTER = _RISK_PRESENT_OR_TOP5_FILTER_factory()

RISK_PRESENT_NO_TOP5_NO_POLICY_DO_EVALUTE_FILTER = sql.and_(
    SurveyTreeItem.type == "risk",
    sql.exists(
        sql.select([Risk.sql_risk_id]).where(
            sql.and_(
                Risk.sql_risk_id == SurveyTreeItem.id,
                sql.not_(Risk.risk_type.in_(["top5", "policy"])),
                sql.not_(Risk.skip_evaluation == True),  # noqa: E712
                Risk.identification == "no",
            )
        )
    ),
)

EVALUATION_FILTER = sql.or_(
    MODULE_WITH_RISK_NO_TOP5_NO_POLICY_DO_EVALUTE_FILTER,
    RISK_PRESENT_NO_TOP5_NO_POLICY_DO_EVALUTE_FILTER,
)

ACTION_PLAN_FILTER = sql.or_(
    MODULE_WITH_RISK_OR_TOP5_FILTER,
    RISK_PRESENT_OR_TOP5_FILTER,
)


def _SKIPPED_MODULE_factory():
    child_node = orm.aliased(SurveyTreeItem)
    sql.exists().where(
        sql.and_(
            SurveyTreeItem.type == "module",
            child_node.session_id == SurveyTreeItem.session_id,
            child_node.skip_children == True,  # noqa: E712
        )
    )


SKIPPED_MODULE = _SKIPPED_MODULE_factory()


def _UNANSWERED_RISKS_FILTER_factory():
    Risk_ = orm.aliased(Risk)
    return sql.and_(
        SurveyTreeItem.type == "risk",
        sql.exists(
            sql.select([Risk_.sql_risk_id]).where(
                sql.and_(
                    Risk_.sql_risk_id == SurveyTreeItem.id,
                    Risk_.identification == None,  # noqa: E711
                )
            )
        ),
    )


UNANSWERED_RISKS_FILTER = _UNANSWERED_RISKS_FILTER_factory()


def _MODULE_WITH_UNANSWERED_RISKS_FILTER_factory():
    child_node = orm.aliased(SurveyTreeItem)
    return sql.and_(
        SurveyTreeItem.type == "module",
        SurveyTreeItem.skip_children == False,  # noqa: E712
        sql.exists(
            sql.select([child_node.id]).where(
                sql.and_(
                    child_node.session_id == SurveyTreeItem.session_id,
                    child_node.id == Risk.sql_risk_id,
                    child_node.type == "risk",
                    Risk.identification is None,
                    child_node.depth > SurveyTreeItem.depth,
                    child_node.path.like(SurveyTreeItem.path + "%"),
                )
            )
        ),
    )


MODULE_WITH_UNANSWERED_RISKS_FILTER = _MODULE_WITH_UNANSWERED_RISKS_FILTER_factory()


def _MODULE_WITH_RISKS_NOT_PRESENT_FILTER_factory():
    child_node = orm.aliased(SurveyTreeItem)
    return sql.and_(
        SurveyTreeItem.type == "module",
        SurveyTreeItem.skip_children == False,  # noqa: E712
        sql.exists(
            sql.select([child_node.id]).where(
                sql.and_(
                    child_node.session_id == SurveyTreeItem.session_id,
                    child_node.id == Risk.sql_risk_id,
                    child_node.type == "risk",
                    Risk.identification == "yes",
                    child_node.depth > SurveyTreeItem.depth,
                    child_node.path.like(SurveyTreeItem.path + "%"),
                )
            )
        ),
    )


MODULE_WITH_RISKS_NOT_PRESENT_FILTER = _MODULE_WITH_RISKS_NOT_PRESENT_FILTER_factory()


def RISK_NOT_PRESENT_FILTER_factory():
    Risk_ = orm.aliased(Risk)
    return sql.and_(
        SurveyTreeItem.type == "risk",
        sql.exists(
            sql.select([Risk_.sql_risk_id]).where(
                sql.and_(
                    Risk_.sql_risk_id == SurveyTreeItem.id,
                    Risk_.identification == "yes",
                )
            )
        ),
    )


RISK_NOT_PRESENT_FILTER = RISK_NOT_PRESENT_FILTER_factory()


def get_current_account():
    """XXX this would be better placed in an api module, but we need to avoid
    circular dependencies.

    :return: The current Account instance if a user can be found,
             otherwise None
    """
    user_id = api.user.get_current().getId()
    if user_id is None:
        return
    try:
        user_id = int(user_id)
    except (ValueError, TypeError):
        # If we pass a user id that can't be transformed to an integer,
        # we get a "DataError (psycopg2.errors.InvalidTextRepresentation)
        # invalid input syntax for type integer.""  When that happens, you get
        # tracebacks like this when another SQL query is done in the same Zope
        # transaction:
        # "psycopg2.errors.InFailedSqlTransaction: current transaction is
        # aborted, commands ignored until end of transaction block"
        # So we prevent this problem by not executing such a query.
        log.warning("Unable to fetch account for non-integer user id: %r", user_id)
        return
    try:
        return Session.query(Account).filter(Account.id == user_id).first()
    except Exception:
        log.warning("Unable to fetch account for user id: %r", user_id)


class DefaultView(BrowserView):
    """Default @@index_html view for the objects in the model."""

    def __call__(self):
        """Somebody called the default view for this object:

        we do not want this to happen so we display a message and
        redirect the user to the session start page
        """
        api.portal.show_message(
            "Wrong URL: %s" % self.request.getURL(), self.request, "warning"
        )
        webhelpers = api.content.get_view("webhelpers", self.context, self.request)
        target = webhelpers.traversed_session.absolute_url() + "/@@start"
        return self.request.response.redirect(target)


@ram.cache(lambda _: "show_timezone_cache_key")
def show_timezone():
    timezone = Session.execute("SHOW TIMEZONE").first()
    return pytz.timezone(timezone[0])


__all__ = [
    "SurveySession",
    "Module",
    "Risk",
    "ActionPlan",
    "SKIPPED_PARENTS",
    "MODULE_WITH_RISK_FILTER",
    "MODULE_WITH_RISK_TOP5_TNO_FILTER",
    "RISK_PRESENT_FILTER",
    "RISK_PRESENT_FILTER_TOP5_TNO_FILTER",
    "get_current_account",
]



from euphorie.client import model
from euphorie.client import update
from euphorie.testing import EuphorieFunctionalTestCase
from plone.dexterity.utils import createContentInContainer
from z3c.saconfig import Session


class TreeTests(EuphorieFunctionalTestCase):
    def createClientSurvey(self):
        self.loginAsPortalOwner()
        self.client = self.portal.client
        self.client.invokeFactory("euphorie.clientcountry", "nl")
        createContentInContainer(
            self.client.nl, "euphorie.sector", checkConstraints=False, title="dining"
        )
        self.sector = self.client.nl.dining
        createContentInContainer(
            self.sector, "euphorie.survey", checkConstraints=False, title="Survey"
        )
        self.survey = self.sector.survey
        return self.survey

    def createSurveySession(self):
        self.sqlsession = Session()
        account = model.Account(loginname="jane", password="secret")
        self.sqlsession.add(account)
        self.session = model.SurveySession(
            title="Session", zodb_path="nl/dining/survey", account=account
        )
        self.sqlsession.add(self.session)
        self.sqlsession.flush()
        return self.session


class GetSurveyTreeTests(TreeTests):
    def testEmpty(self):
        survey = self.createClientSurvey()
        self.assertEqual(update.getSurveyTree(survey), [])

    def testSingleModule(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        self.assertEqual(
            update.getSurveyTree(survey),
            [
                {
                    "optional": False,
                    "zodb_path": "1",
                    "type": "module",
                    "has_description": False,
                    "always_present": False,
                    "risk_type": None,
                }
            ],
        )

    def testModuleAndRisk(self):
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        survey["1"].invokeFactory("euphorie.risk", "2")
        self.assertEqual(
            update.getSurveyTree(survey),
            [
                {
                    "optional": False,
                    "zodb_path": "1",
                    "type": "module",
                    "has_description": False,
                    "always_present": False,
                    "risk_type": None,
                },
                {
                    "optional": False,
                    "zodb_path": "1/2",
                    "type": "risk",
                    "has_description": False,
                    "always_present": False,
                    "risk_type": "risk",
                },
            ],
        )


class GetSessionTreeTests(TreeTests):
    def testEmpty(self):
        session = self.createSurveySession()
        self.assertEqual(update.getSessionTree(session), [])

    def testSingleModule(self):
        session = self.createSurveySession()
        session.addChild(model.Module(title="Root", module_id="1", zodb_path="1"))
        tree = update.getSessionTree(session)
        self.assertEqual(len(tree), 1)
        self.assertEqual(tree[0].zodb_path, "1")
        self.assertEqual(tree[0].type, "module")
        self.assertEqual(tree[0].path, "001")

    def testModuleAndRisk(self):
        session = self.createSurveySession()
        root = model.Module(title="Root", module_id="1", zodb_path="1")
        session.addChild(root)
        root.addChild(
            model.Risk(
                title="Risk 1",
                risk_id="1",
                zodb_path="1/1",
                type="risk",
                identification="no",
            )
        )
        tree = update.getSessionTree(session)
        self.assertEqual(len(tree), 2)
        self.assertEqual(tree[0].zodb_path, "1")
        self.assertEqual(tree[0].type, "module")
        self.assertEqual(tree[0].path, "001")
        self.assertEqual(tree[1].zodb_path, "1/1")
        self.assertEqual(tree[1].type, "risk")
        self.assertEqual(tree[1].path, "001001")


class TreeChangesTests(TreeTests):
    def test_nothing_changes(self):
        session = self.createSurveySession()
        session_module = model.Module(
            title="Root",
            module_id="1",
            zodb_path="1",
            skip_children=False,
            has_description=False,
        )
        session.addChild(session_module)
        session_module.addChild(
            model.Risk(
                title="Risk 1",
                risk_id="2",
                zodb_path="1/2",
                type="risk",
                identification="no",
            )
        )
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        survey["1"].invokeFactory("euphorie.risk", "2")
        self.assertEqual(update.treeChanges(session, survey), set())

    def test_no_changes_with_repeated_profile(self):
        session = self.createSurveySession()
        for i in range(2):
            session_module = model.Module(
                title="Root",
                module_id="1",
                zodb_path="1",
                skip_children=False,
                profile_index=i,
            )
            session.addChild(session_module)
            session_module.addChild(
                model.Risk(
                    title="Risk 1",
                    risk_id="2",
                    zodb_path="1/2",
                    type="risk",
                    identification="no",
                )
            )

        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        survey["1"].invokeFactory("euphorie.risk", "2")
        self.assertEqual(update.treeChanges(session, survey), set())

    def testAddNewModule(self):
        session = self.createSurveySession()
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        changes = update.treeChanges(session, survey)
        self.assertEqual(changes, {("1", "module", "add")})

    def test_add_new_risk(self):
        session = self.createSurveySession()
        session.addChild(
            model.Module(
                title="Root", module_id="1", zodb_path="1", has_description=False
            )
        )
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        survey["1"].invokeFactory("euphorie.risk", "2")
        changes = update.treeChanges(session, survey)
        self.assertEqual(changes, {("1/2", "risk", "add")})

    def testProfileActsAsModule(self):
        session = self.createSurveySession()
        session.addChild(model.Module(title="Root", module_id="1", zodb_path="1"))
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.profilequestion", "1")
        changes = update.treeChanges(session, survey)
        self.assertEqual(changes, set())

    def testRemoveModule(self):
        session = self.createSurveySession()
        session.addChild(model.Module(title="Root", module_id="1", zodb_path="1"))
        survey = self.createClientSurvey()
        changes = update.treeChanges(session, survey)
        self.assertEqual(changes, {("1", "module", "remove")})

    def test_remove_risk(self):
        session = self.createSurveySession()
        session.addChild(
            model.Module(
                title="Root", module_id="1", zodb_path="1", has_description=False
            )
        )
        session.addChild(
            model.Risk(
                title="Risk 1",
                risk_id="1",
                zodb_path="1/1",
                type="risk",
                identification="no",
            )
        )
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        changes = update.treeChanges(session, survey)
        self.assertEqual(changes, {("1/1", "risk", "remove")})

    def test_module_lost_description(self):
        session = self.createSurveySession()
        session_module = model.Module(
            title="Root",
            module_id="1",
            zodb_path="1",
            skip_children=False,
            has_description=True,
        )
        session.addChild(session_module)
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        module = survey["1"]
        module.description = "<p></p>"
        changes = update.treeChanges(session, survey)
        self.assertEqual(changes, {("1", "module", "modified")})

    def test_module_gained_description(self):
        session = self.createSurveySession()
        session_module = model.Module(
            title="Root",
            module_id="1",
            zodb_path="1",
            skip_children=False,
            has_description=False,
        )
        session.addChild(session_module)
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        module = survey["1"]
        module.description = "<p>Hello</p>"
        changes = update.treeChanges(session, survey)
        self.assertEqual(changes, {("1", "module", "modified")})

    def testModuleMadeRequired(self):
        session = self.createSurveySession()
        session_module = model.Module(
            title="Root", module_id="1", zodb_path="1", skip_children=True
        )
        session.addChild(session_module)
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        module = survey["1"]
        module.optional = False
        changes = update.treeChanges(session, survey)
        self.assertEqual(changes, {("1", "module", "modified")})

    def test_module_made_optional(self):
        """
        Note: an optional module is always considered to have a description, so
        that is never gets skipped.
        See https://github.com/euphorie/Euphorie/commit/c277d05
        """
        session = self.createSurveySession()
        session_module = model.Module(
            title="Root",
            module_id="1",
            zodb_path="1",
            has_description=True,
            skip_children=False,
        )
        session.addChild(session_module)
        survey = self.createClientSurvey()
        survey.invokeFactory("euphorie.module", "1")
        module = survey["1"]
        module.optional = True
        changes = update.treeChanges(session, survey)
        self.assertEqual(changes, set())


class MeasureChangeTests(TreeTests):
    def test_update_measure_types(self):
        survey = self.createClientSurvey()
        session = self.createSurveySession()
        survey.invokeFactory("euphorie.module", "1")
        survey["1"].invokeFactory("euphorie.risk", "2")
        solution_id = survey["1"]["2"].invokeFactory("euphorie.solution", "test")

        root = session.addChild(
            model.Module(title="Root", module_id="1", zodb_path="1")
        )
        risk = root.addChild(
            model.Risk(
                title="Risk 1",
                risk_id="2",
                zodb_path="1/2",
                type="risk",
                identification="no",
            )
        )
        solution = model.ActionPlan(
            solution_id=solution_id,
            action_plan="This is the plan",
            plan_type="in_place_standard",
            risk=risk,
        )

        session.update_measure_types(survey)
        self.assertEqual(solution.plan_type, "in_place_standard")

        risk_zodb = survey["1"]["2"]
        risk_zodb.manage_delObjects([solution_id])
        if hasattr(risk_zodb, "_memojito_"):
            risk_zodb._memojito_.clear()
        session.update_measure_types(survey)
        self.assertEqual(solution.plan_type, "in_place_custom")



"""
Utils
-----

Helper functions.
"""

from .. import MessageFactory as _
from email.header import Header
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from euphorie.client import model
from euphorie.content.utils import StripMarkup
from plone import api
from Products.CMFCore.utils import getToolByName
from sqlalchemy import sql
from z3c.saconfig import Session
from zope.i18nmessageid import MessageFactory

import email.utils as emailutils
import logging
import random
import threading


locals = threading.local()
log = logging.getLogger(__name__)

pl_message = MessageFactory("plonelocales")


def setRequest(request):
    locals.request = request


def getRequest():
    return getattr(locals, "request", None)


def getSecret():
    try:
        site = api.portal.get()
    except api.exc.CannotGetPortalError:
        site = None
    return getattr(site, "euphorie_secret", "secret")


def randomString(length=16):
    """Return 32 bytes of random data.

    Only characters which do not require special escaping in HTML or
    URLs are generated.
    """
    safe_characters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890-"
    return "".join(random.choice(safe_characters) for idx in range(length))


def get_translated_custom_risks_title(request):
    return api.portal.translate(_("label_custom_risks", default="Custom risks"))


def HasText(html):
    """Determine if a HTML fragment contains text."""
    if not html:
        return False
    text = StripMarkup(html).replace(" ", "").replace("&nbsp;", "")
    return bool(text)


def CreateEmailTo(sender_name, sender_email, recipient, subject, body):
    mail = MIMEMultipart("alternative")
    mail["From"] = emailutils.formataddr((sender_name, sender_email))
    mail["To"] = recipient
    mail["Subject"] = Header(subject.encode("utf-8"), "utf-8")
    mail["Date"] = emailutils.formatdate(localtime=True)
    mail.set_param("charset", "utf-8")
    if isinstance(body, str):
        mail.attach(MIMEText(body.encode("utf-8"), "plain", "utf-8"))
    else:
        mail.attach(MIMEText(body))

    return mail


def setLanguage(request, context, lang=None):
    """Switch Plone to another language.

    If no language is given via the `lang` parameter the language is
    taken from a `language` request parameter. If a dialect was chosen
    but is not available the main language is used instead. If the main
    language is also unavailable switch back to English.
    """
    if lang is None:
        lang = request.form.get("language")
    if not lang:
        return

    lang = lang.lower()
    lt = getToolByName(context, "portal_languages")
    res = lt.setLanguageCookie(lang=lang, request=request)
    if res is None and "-" in lang:
        lang = lang.split("-")[0]
        res = lt.setLanguageCookie(lang=lang, request=request)
        if res is None:
            log.warning("Failed to switch language to %r", lang)
            lt.setLanguageCookie(lang="en", request=request)
            lang = "en"

    # In addition to setting the cookie also update the PTS language.
    # This effectively switches Plone over to the new language without
    # requiring a new HTTP request.
    request["LANGUAGE"] = lang
    binding = request.get("LANGUAGE_TOOL", None)
    if binding is not None:
        binding.LANGUAGE = lang


def remove_empty_modules(nodes):
    """Takes a list of modules and risks.

    Removes modules that don't have any risks in them.
    Modules with submodules (with risks) must however be kept.

    How it works:
    -------------
    Use the 'grow' method to create a tree datastructure that
    mirrors the actual layout of modules and risks.

    Then 'prune' it by removing all branches that end in modules.

    Lastly flatten the tree back into a list and use it to filter the
    original list.
    """
    tree = {}
    ids = []

    def grow(tree, nodes):
        for i in range(0, len(nodes)):
            node = nodes[i]
            inserted = False
            for k in tree.keys():
                if node.path.startswith(k[0]):
                    if tree[k]:
                        grow(tree[k], [node])
                    else:
                        tree[k] = {(node.path, node.type, node.id): {}}
                    inserted = True
                    break
            if not inserted:
                tree[(node.path, node.type, node.id)] = {}

    def prune(tree):
        for k in list(tree):
            if tree[k]:
                prune(tree[k])

            if not tree[k] and k[1] == "module":
                del tree[k]

    def flatten(tree):
        for k in tree.keys():
            ids.append(k[2])
            flatten(tree[k])

    grow(tree, nodes)
    prune(tree)
    flatten(tree)
    return [n for n in nodes if n.id in ids]


def get_unactioned_nodes(ls, filter_for_measures=False):
    """Takes a list of modules and risks and removes all risks that have been
    actioned (i.e has at least one valid action plan). Also remove all modules
    that have lost all their risks in the process.

    See https://syslab.com/proj/issues/2885
    """
    unactioned = []
    for n in ls:
        if n.type == "module":
            unactioned.append(n)

        elif n.type == "risk":
            action_plans = n.standard_measures + n.custom_measures
            if not action_plans:
                if filter_for_measures:
                    if not (n.in_place_standard + n.in_place_custom):
                        unactioned.append(n)
                else:
                    unactioned.append(n)
            else:
                # It's possible that there is an action plan object, but
                # that it's not yet fully populated
                if action_plans[0] is None or action_plans[0].action is None:
                    unactioned.append(n)

    return remove_empty_modules(unactioned)


def get_actioned_nodes(ls):
    """Takes a list of modules and risks and removes all risks that are *not*
    actioned (i.e does not have at least one valid action plan) Also remove all
    modules that have lost all their risks in the process.

    See https://syslab.com/proj/issues/2885
    """
    actioned = []
    for n in ls:
        if n.type == "module":
            actioned.append(n)

        if n.type == "risk":
            action_plans = n.standard_measures + n.custom_measures
            if len(action_plans):
                # It's possible that there is an action plan object, but
                # it's not yet fully populated
                plans = [p.action for p in action_plans]
                if plans[0] is not None:
                    actioned.append(n)

    return remove_empty_modules(actioned)


def get_unanswered_nodes(session):
    query = (
        Session()
        .query(model.SurveyTreeItem)
        .filter(
            sql.and_(
                model.SurveyTreeItem.session == session,
                sql.or_(
                    model.MODULE_WITH_UNANSWERED_RISKS_FILTER,
                    model.UNANSWERED_RISKS_FILTER,
                ),
                sql.not_(model.SKIPPED_PARENTS),
            )
        )
        .order_by(model.SurveyTreeItem.path)
    )
    return query.all()


def get_risk_not_present_nodes(session):
    query = (
        Session()
        .query(model.SurveyTreeItem)
        .filter(
            sql.and_(
                model.SurveyTreeItem.session == session,
                sql.or_(
                    model.SKIPPED_PARENTS,
                    model.MODULE_WITH_RISKS_NOT_PRESENT_FILTER,
                    model.RISK_NOT_PRESENT_FILTER,
                    model.SKIPPED_MODULE,
                ),
            )
        )
        .order_by(model.SurveyTreeItem.path)
    )
    return query.all()


def get_italian_risk_not_present_nodes(session):
    query = (
        Session()
        .query(model.SurveyTreeItem)
        .filter(
            sql.and_(
                model.SurveyTreeItem.session == session,
                sql.or_(
                    model.SKIPPED_PARENTS,
                    model.MODULE_WITH_RISKS_NOT_PRESENT_FILTER,
                    model.SKIPPED_MODULE,
                    model.UNANSWERED_RISKS_FILTER,
                ),
            )
        )
        .order_by(model.SurveyTreeItem.path)
    )
    return query.all()



from euphorie.client import model
from euphorie.client import navigation
from euphorie.client.browser.risk import ActionPlanView
from euphorie.testing import EuphorieIntegrationTestCase

import unittest


def createSurvey():
    session = model.Session()
    account = model.Account(loginname="jane", password="secret")
    session.add(account)
    survey = model.SurveySession(
        title="Session",
        zodb_path="survey",
        account=account,
    )
    session.add(survey)
    return (session, survey)


class MockRequest:
    def __init__(self, agent=None):
        self.__headers = {}
        if agent is not None:
            self.__headers["User-Agent"] = agent

    def get_header(self, key, default):
        return self.__headers.get(key, default)


class FindNextQuestionTests(EuphorieIntegrationTestCase):
    def testSingleQuestion(self):
        (session, survey) = createSurvey()
        child = model.Risk(title="Risk", risk_id="1", zodb_path="1")
        survey.addChild(child)
        self.assertTrue(navigation.FindNextQuestion(child, survey) is None)

    def test_Question_at_same_level_as_module(self):
        (session, survey) = createSurvey()
        session.add(survey)
        child = model.Risk(title="Risk", risk_id="1", zodb_path="1")
        survey.addChild(child)
        sister = model.Risk(title="Risk", risk_id="2", zodb_path="2")
        survey.addChild(sister)
        self.assertTrue(navigation.FindNextQuestion(child, survey) is sister)

    def testQuestionIsNextModule(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(title="Module 1", module_id="1", zodb_path="1")
        survey.addChild(mod1)
        q1 = model.Risk(title="Risk 1", risk_id="1", zodb_path="1/2")
        mod1.addChild(q1)
        mod2 = model.Module(
            title="Module 2", module_id="2", zodb_path="2", has_description=True
        )
        survey.addChild(mod2)
        self.assertTrue(navigation.FindNextQuestion(q1, survey) is mod2)

    def testSkipChildren(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(
            title="Module 1", module_id="1", zodb_path="1", skip_children=True
        )
        survey.addChild(mod1)
        q1 = model.Risk(
            title="Risk 1", risk_id="1", zodb_path="1/1", has_description=True
        )
        mod1.addChild(q1)
        mod2 = model.Module(
            title="Module 2", module_id="2", zodb_path="2", has_description=True
        )
        survey.addChild(mod2)
        self.assertTrue(navigation.FindNextQuestion(mod1, survey) is mod2)

    def test_ignore_module_without_description(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(title="Module 1", module_id="1", zodb_path="1")
        survey.addChild(mod1)
        q1 = model.Risk(title="Risk 1", risk_id="1", zodb_path="1/1")
        mod1.addChild(q1)
        mod2 = model.Module(
            title="Module 2", module_id="2", zodb_path="2", has_description=False
        )
        survey.addChild(mod2)
        mod3 = model.Module(
            title="Module 3", module_id="3", zodb_path="3", has_description=True
        )
        survey.addChild(mod3)
        self.assertTrue(navigation.FindNextQuestion(q1, survey) is mod3)


class FindPreviousQuestionTests(EuphorieIntegrationTestCase):
    def testSingleQuestion(self):
        (session, survey) = createSurvey()
        child = model.Risk(title="Risk", risk_id="1", zodb_path="1")
        survey.addChild(child)
        self.assertTrue(navigation.FindPreviousQuestion(child, survey) is None)

    def testQuestionAtSameModule(self):
        (session, survey) = createSurvey()
        child = model.Risk(title="Risk 1", risk_id="1", zodb_path="1")
        survey.addChild(child)
        sister = model.Risk(title="Risk 2", risk_id="2", zodb_path="2")
        survey.addChild(sister)
        self.assertTrue(navigation.FindPreviousQuestion(sister, survey) is child)

    def testQuestionAtPreviousModule(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(title="Module 1", module_id="1", zodb_path="1")
        survey.addChild(mod1)
        q1 = model.Risk(title="Risk 1", risk_id="1", zodb_path="1/1")
        mod1.addChild(q1)
        mod2 = model.Module(title="Module 2", module_id="2", zodb_path="2")
        survey.addChild(mod2)
        self.assertTrue(navigation.FindPreviousQuestion(mod2, survey) is q1)

    def testQuestionAtPreviousModuleWithSkippedChildren(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(
            title="Module 1",
            module_id="1",
            zodb_path="1",
            has_description=True,
            skip_children=True,
        )
        survey.addChild(mod1)
        q1 = model.Risk(title="Risk 1", risk_id="1", zodb_path="1/1")
        mod1.addChild(q1)
        mod2 = model.Module(title="Module 2", module_id="2", zodb_path="2")
        survey.addChild(mod2)
        self.assertTrue(navigation.FindPreviousQuestion(mod2, survey) is mod1)

    def test_skip_module_without_description(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(
            title="Module 1", module_id="1", zodb_path="1", has_description=True
        )
        survey.addChild(mod1)
        q1 = model.Risk(title="Risk 1", risk_id="1", zodb_path="1/1")
        mod1.addChild(q1)
        mod2 = model.Module(
            title="Module 2", module_id="2", zodb_path="2", has_description=False
        )
        survey.addChild(mod2)
        mod3 = model.Module(title="Module 3", module_id="3", zodb_path="3")
        survey.addChild(mod3)
        self.assertTrue(navigation.FindPreviousQuestion(mod3, survey) is q1)


class ActionPlanNavigationTests(EuphorieIntegrationTestCase):
    """Test if the filter determining which modules and risks to show during
    the action plan phase are correct."""

    def filter(self):
        return ActionPlanView.question_filter

    def testSkipModuleWithoutRisks(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(
            title="Module 1", module_id="1", zodb_path="1", skip_children=False
        )
        survey.addChild(mod1)
        mod11 = model.Module(
            title="Module 1.1", module_id="11", zodb_path="1/1", skip_children=False
        )
        mod1.addChild(mod11)
        self.assertEqual(navigation.FindNextQuestion(mod1, survey, self.filter()), None)

    def testSkipModuleIfNoRisksPresent(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(
            title="Module 1", module_id="1", zodb_path="1", skip_children=False
        )
        survey.addChild(mod1)
        mod11 = model.Module(
            title="Module 1.1", module_id="11", zodb_path="1/1", skip_children=False
        )
        mod1.addChild(mod11)
        q111 = model.Risk(
            title="Risk 1.1.1", risk_id="1", zodb_path="1/1/1", identification="yes"
        )
        mod11.addChild(q111)
        self.assertEqual(navigation.FindNextQuestion(mod1, survey, self.filter()), None)

    def testShowModuleWithTop5RiskEvenIfNotPresent(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(
            title="Module 1", module_id="1", zodb_path="1", skip_children=False
        )
        survey.addChild(mod1)
        mod11 = model.Module(
            title="Module 1.1",
            module_id="11",
            zodb_path="1/1",
            skip_children=False,
            has_description=True,
        )
        mod1.addChild(mod11)
        q111 = model.Risk(
            title="Risk 1.1.1",
            risk_id="1",
            zodb_path="1/1/1",
            identification="yes",
            risk_type="top5",
        )
        mod11.addChild(q111)
        self.assertTrue(
            navigation.FindNextQuestion(mod1, survey, self.filter()) is mod11
        )

    def testSkipRiskIfNotPresent(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(
            title="Module 1", module_id="1", zodb_path="1", skip_children=False
        )
        survey.addChild(mod1)
        q11 = model.Risk(
            title="Risk 1.1", risk_id="1", zodb_path="1/1", identification="yes"
        )
        mod1.addChild(q11)
        self.assertEqual(navigation.FindNextQuestion(mod1, survey, self.filter()), None)

    def testShowTop5RiskEvenIfNotPresent(self):
        (session, survey) = createSurvey()
        mod1 = model.Module(
            title="Module 1", module_id="1", zodb_path="1", skip_children=False
        )
        survey.addChild(mod1)
        q11 = model.Risk(
            title="Risk 1.1",
            risk_id="1",
            zodb_path="1/1",
            identification="yes",
            risk_type="top5",
        )
        mod1.addChild(q11)
        self.assertTrue(navigation.FindNextQuestion(mod1, survey, self.filter()) is q11)


class GetTreeDataTests(EuphorieIntegrationTestCase):
    def createSqlData(self):
        self.request = MockRequest()
        (self.session, self.survey) = createSurvey()
        self.survey.restrictedTraverse = lambda x: None
        self.request.survey = self.survey
        self.survey.absolute_url = lambda self=None: "http://nohost"
        self.session.flush()
        self.mod1 = self.survey.addChild(
            model.Module(title="module 1", module_id="1", zodb_path="a")
        )
        self.q1 = self.mod1.addChild(
            model.Risk(title="question 1", risk_id="1", zodb_path="a/b")
        )
        self.session.flush()

    def testMinimalTree(self):
        self.createSqlData()
        data = navigation.getTreeData(self.request, self.q1, survey=self.survey)
        self.assertTrue(isinstance(data, dict))
        children = data["children"]
        self.assertEqual(len(children), 1)
        mod1 = children[0]
        self.assertEqual(mod1["id"], self.mod1.id)
        self.assertEqual(mod1["type"], "module")
        self.assertEqual(mod1["number"], "1")
        self.assertEqual(mod1["title"], "module 1")
        self.assertEqual(mod1["current"], False)
        self.assertEqual(mod1["current_parent"], True)
        self.assertEqual(mod1["leaf_module"], True)
        self.assertEqual(mod1["active"], True)
        self.assertEqual(mod1["class"], "active current_parent")
        self.assertEqual(mod1["url"], "http://nohost/1/@@identification")
        self.assertEqual(len(mod1["children"]), 1)
        q1 = mod1["children"][0]
        self.assertEqual(q1["id"], self.q1.id)
        self.assertEqual(q1["type"], "risk")
        self.assertEqual(q1["number"], "1.1")
        self.assertEqual(q1["title"], "question 1")
        self.assertEqual(q1["current"], True)
        self.assertEqual(q1["active"], False)
        self.assertEqual(q1["class"], "current")
        self.assertEqual(q1["url"], "http://nohost/1/1/@@identification")
        self.assertEqual(len(q1["children"]), 0)

    def testIncludeRisksChildrenOfModule(self):
        self.createSqlData()
        data = navigation.getTreeData(self.request, self.mod1, survey=self.survey)
        mod1_data = data["children"][0]
        self.assertEqual(len(mod1_data["children"]), 1)
        self.assertEqual(mod1_data["leaf_module"], True)

    def testIncludeRisksChildrenOfModuleUnlessSkipped(self):
        self.createSqlData()
        self.mod1.skip_children = True
        data = navigation.getTreeData(self.request, self.mod1, survey=self.survey)
        mod1_data = data["children"][0]
        self.assertEqual(len(mod1_data["children"]), 0)
        self.assertEqual(mod1_data["leaf_module"], False)

    def test_list_sibling_modules_of_parent_if_risk(self):
        self.createSqlData()
        self.mod1.removeChildren()
        mod11 = self.mod1.addChild(
            model.Module(title="module 1.1", module_id="11", zodb_path="a/a")
        )
        mod12 = self.mod1.addChild(
            model.Module(title="module 1.2", module_id="12", zodb_path="a/b")
        )
        q111 = mod11.addChild(
            model.Risk(title="question 1.1.1", risk_id="111", zodb_path="a/b/c")
        )
        data = navigation.getTreeData(self.request, q111, survey=self.survey)
        mod1_data = data["children"][0]
        self.assertEqual(len(mod1_data["children"]), 2)
        self.assertEqual(mod1_data["children"][0]["id"], mod11.id)
        self.assertEqual(mod1_data["children"][0]["current_parent"], True)
        self.assertEqual(mod1_data["children"][0]["current"], False)
        self.assertEqual(mod1_data["children"][0]["leaf_module"], True)
        self.assertEqual(mod1_data["children"][0]["active"], True)
        self.assertEqual(mod1_data["children"][1]["id"], mod12.id)
        self.assertEqual(mod1_data["children"][1]["current_parent"], False)
        self.assertEqual(mod1_data["children"][1]["current"], False)
        self.assertEqual(mod1_data["children"][1]["active"], False)

    def testIgnoreSiblingQuestions(self):
        self.createSqlData()
        self.mod1.removeChildren()
        mod11 = self.mod1.addChild(
            model.Module(title="module 1.1", module_id="11", zodb_path="a/a")
        )
        q111 = mod11.addChild(
            model.Risk(title="question 1.1.1", risk_id="111", zodb_path="a/a/a")
        )
        self.mod1.addChild(
            model.Risk(title="question 1.2", risk_id="12", zodb_path="a/b")
        )
        data = navigation.getTreeData(self.request, q111, survey=self.survey)
        mod1_data = data["children"][0]
        self.assertEqual(len(mod1_data["children"]), 1)
        self.assertEqual(mod1_data["children"][0]["id"], mod11.id)

    def testListRootSiblingModules(self):
        self.createSqlData()
        self.mod1.removeChildren()
        mod11 = self.mod1.addChild(
            model.Module(title="module 1.1", module_id="11", zodb_path="a/a")
        )
        q111 = mod11.addChild(
            model.Risk(title="question 1.1.1", risk_id="111", zodb_path="a/a/a")
        )
        mod2 = self.survey.addChild(
            model.Module(title="module 2", module_id="2", zodb_path="b")
        )
        data = navigation.getTreeData(self.request, q111, survey=self.survey)
        self.assertEqual(len(data["children"]), 2)
        self.assertEqual(data["children"][0]["id"], self.mod1.id)
        self.assertEqual(data["children"][0]["current"], False)
        self.assertEqual(data["children"][0]["active"], True)
        self.assertEqual(data["children"][1]["id"], mod2.id)
        self.assertEqual(data["children"][1]["current"], False)
        self.assertEqual(data["children"][1]["active"], False)


class FirstTests(unittest.TestCase):
    def testNonIterableRaisesTypeError(self):
        self.assertRaises(TypeError, navigation.first, lambda x: x, None)

    def testEmptyList(self):
        self.assertEqual(navigation.first(lambda x: x, []), None)

    def testNoMatchInList(self):
        self.assertEqual(navigation.first(lambda x: x, [False]), None)

    def testMatch(self):
        self.assertEqual(navigation.first(lambda x: x, [False, True]), True)

    def testFirstMatch(self):
        self.assertEqual(navigation.first(lambda x: x, [1, 2]), 1)

    def testGenerator(self):
        self.assertEqual(navigation.first(lambda x: x, range(5)), 1)



from euphorie.client import model
from euphorie.client.model import SurveySession
from euphorie.client.tests.utils import addAccount
from euphorie.client.tests.utils import addSurvey
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api


class TestHistoryPopup(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        with api.env.adopt_user("admin"):
            addSurvey(self.portal, BASIC_SURVEY)
        self.account = addAccount(password="secret")
        self.survey = self.portal.client.nl.ict["software-development"]
        survey_session = SurveySession(
            title="Dummy session",
            zodb_path="nl/ict/software-development",
            account=self.account,
            company=model.Company(country="nl", employees="1-9", referer="other"),
        )
        model.Session.add(survey_session)

    def test_history_popup(self):
        with api.env.adopt_user(user=self.account):
            traversed_session = self.survey.restrictedTraverse("++session++1")
            with self._get_view("history_popup", traversed_session) as view:
                # By default there is no event registered but the history starts
                # using the creation date of the session
                self.assertEqual(len(view.events), 0)
                self.assertEqual(len(view.items), 1)
                item = view.items[0]
                self.assertEqual(item.email, "jane@example.com")
                self.assertEqual(item.message, "started assessment")



"""
Country
-------

Is the container for sectors of that country, plus the country manager user
accounts.

https://admin.oiraproject.eu/sectors/eu
"""

from .. import MessageFactory as _
from plone.app.dexterity.behaviors.metadata import IBasic
from plone.autoform import directives
from plone.dexterity.content import Container
from plone.supermodel import model
from plonetheme.nuplone.z3cform.directives import depends
from z3c.form.browser.checkbox import CheckBoxFieldWidget
from zope import schema
from zope.interface import implementer
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


class ICountry(model.Schema, IBasic):
    """Country grouping in the online client."""

    country_type = schema.Choice(
        title=_("Country grouping"),
        vocabulary=SimpleVocabulary(
            [
                SimpleTerm("region", title=_("Region")),
                SimpleTerm("eu-member", title=_("EU member state")),
                SimpleTerm("efta", title=_("EFTA country")),
                SimpleTerm("candidate-eu", title=_("Candidate country")),
                SimpleTerm(
                    "potential-candidate-eu", title=_("Potential candidate country")
                ),
            ]
        ),
        default="eu-member",
        required=True,
    )

    directives.widget(risk_default_collapsible_sections=CheckBoxFieldWidget)
    risk_default_collapsible_sections = schema.List(
        title=_(
            "label__risk_default_collapsible_sections",
            "Expanded sections on risk page",
        ),
        description=_(
            "help__risk_default_collapsible_sections",
            "Define, which information sections should be open by "
            "default on a risk identification page. Sections not checked "
            "will be shown intially in collapsed mode, but the user can always "
            "open those sections with a click.",
        ),
        value_type=schema.Choice(
            vocabulary=SimpleVocabulary(
                [
                    SimpleTerm(
                        "collapsible_section_information", title=_("Information")
                    ),
                    SimpleTerm(
                        "collapsible_section_resources",
                        title=_("Resources: Legal references and attachments"),
                    ),
                    SimpleTerm("collapsible_section_comments", title=_("Comments")),
                ]
            )
        ),
        default=["collapsible_section_information"],
        required=False,
    )

    directives.widget(default_reports=CheckBoxFieldWidget)
    default_reports = schema.List(
        title=_("label__default_reports", "Available reports"),
        description=_(
            "help__default_reports",
            "Define which reports are offered to the user on the Report page.",
        ),
        value_type=schema.Choice(
            vocabulary=SimpleVocabulary(
                [
                    SimpleTerm(
                        "report_compact", title=_("Compact report (Word document)")
                    ),
                    SimpleTerm("report_full", title=_("Full report (Word document)")),
                    SimpleTerm(
                        "report_action_plan",
                        title=_("Action plan (Excel spreadsheet)"),
                    ),
                    SimpleTerm(
                        "report_overview_risks", title=_("Overview of risks (PDF)")
                    ),
                    SimpleTerm(
                        "report_overview_measures",
                        title=_("Overview of measures (PDF)"),
                    ),
                ]
            )
        ),
        default=["report_full", "report_action_plan", "report_overview_risks"],
        required=False,
    )

    depends("compact_report_options", "default_reports", "==", "report_compact")
    directives.widget(compact_report_options=CheckBoxFieldWidget)
    compact_report_options = schema.List(
        title=_("label_compact_report_options", "Compact report options"),
        description=_(
            "help_compact_report_options",
            "Select which optional elements are used in the compact report",
        ),
        value_type=schema.Choice(
            vocabulary=SimpleVocabulary(
                [
                    SimpleTerm("description_of_risks", title=_("Description of risks")),
                    SimpleTerm("comments", title=_("Comments")),
                    SimpleTerm(
                        "description_of_measures", title=_("Description of measures")
                    ),
                    SimpleTerm(
                        "measure_responsible_date",
                        title=_("Measure responsible and date"),
                    ),
                    SimpleTerm(
                        "resources_legal_references",
                        title=_("Legal and policy references"),
                    ),
                ]
            )
        ),
        default=[
            "description_of_risks",
            "comments",
            "description_of_measures",
            "measure_responsible_date",
        ],
        required=False,
    )

    enable_web_training = schema.Bool(
        title=_("label_enable_web_training", default="Enable Web Based Training?"),
        description=_(
            "help_enable_web_training_country",
            default="If this option is activated, an online training can be enabled on "
            "OiRA tools in this country.",
        ),
        required=False,
        default=True,
    )

    enable_consultancy = schema.Bool(
        title=_("label_enable_consultancy", default="Enable Consultancy?"),
        description=_(
            "help_enable_consultancy_country",
            default="If this option is activated, assessments can be reviewed and "
            "validated by consultants.",
        ),
        required=False,
        default=False,
    )


@implementer(ICountry)
class Country(Container):
    """A country folder."""

    def _canCopy(self, op=0):
        """Tell Zope2 that this object can not be copied."""
        return False



# from cStringIO import StringIO
# from euphorie.client.report import HtmlToRtf
# from euphorie.client.report import IdentificationReport
# from euphorie.content.risk import Risk
from euphorie.client import model
from euphorie.client import utils
from euphorie.client.adapters.session_traversal import TraversedSurveySession
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.client.model import SurveySession
from euphorie.client.tests.test_model import createSurvey
from euphorie.client.tests.utils import addAccount
from euphorie.testing import EuphorieIntegrationTestCase
from ExtensionClass import Base
from plone import api
from plone.app.testing.interfaces import SITE_OWNER_NAME
from unittest import mock
from urllib.parse import quote
from z3c.saconfig import Session
from zope.interface import alsoProvides

import datetime


# import unittest


class ReportIntegrationTests(EuphorieIntegrationTestCase):
    def create_session(self):
        with api.env.adopt_user(SITE_OWNER_NAME):
            api.content.create(
                container=self.portal.sectors, type="euphorie.country", id="eu"
            )
            client_country = api.content.create(
                container=self.portal.client, type="euphorie.clientcountry", id="eu"
            )
            client_sector = api.content.create(
                container=client_country, type="euphorie.clientsector", id="sector"
            )
            api.content.create(
                container=client_sector, type="euphorie.survey", id="survey"
            )

        sqlsession = Session()
        account = model.Account(loginname="jane", password="secret")
        sqlsession.add(account)
        session = model.SurveySession(
            title="Session", zodb_path="eu/sector/survey", account=account
        )
        sqlsession.add(session)
        sqlsession.flush()

        return session

    def test_default_reports(self):
        self.create_session()
        traversed_session = self.portal.client.eu.sector.survey.restrictedTraverse(
            "++session++1"
        )

        with self._get_view("report_view", traversed_session) as view:
            # default default sections
            self.assertEqual(
                view.default_reports,
                ["report_full", "report_action_plan", "report_overview_risks"],
            )

        # customized default sections
        view.webhelpers.content_country_obj.default_reports = [
            "report_overview_measures",
        ]
        with self._get_view("report_view", traversed_session) as view:
            self.assertEqual(
                view.default_reports,
                ["report_overview_measures"],
            )


# XXX Change these tests to test client.docx.views.IdentificationReportDocxView instead

# class IdentificationReportTests(unittest.TestCase):

#     def IdentificationReport(self, *a, **kw):
#         return IdentificationReport(*a, **kw)

#     def test_title_not_a_risk(self):
#         node = mock.Mock()
#         node.type = 'module'
#         node.title = u'My title'
#         view = self.IdentificationReport(None, None)
#         self.assertEqual(view.title(node, None), u'My title')

#     def test_title_unanswered_risk(self):
#         node = mock.Mock()
#         node.type = 'risk'
#         node.identification = None
#         node.title = u'My title'
#         view = self.IdentificationReport(None, None)
#         self.assertEqual(view.title(node, None), u'My title')

#     def test_title_empty_problem_description(self):
#         node = mock.Mock()
#         node.type = 'risk'
#         node.identification = u'no'
#         node.title = u'My title'
#         zodb_node = mock.Mock()
#         zodb_node.problem_description = u'   '
#         view = self.IdentificationReport(None, None)
#         self.assertEqual(view.title(node, zodb_node), u'My title')

#     def test_title_risk_present_and_with_problem_description(self):
#         node = mock.Mock()
#         node.type = 'risk'
#         node.identification = u'no'
#         node.title = u'My title'
#         zodb_node = mock.Mock()
#         zodb_node.problem_description = u'Bad situation'
#         view = self.IdentificationReport(None, None)
#         self.assertEqual(view.title(node, zodb_node), u'Bad situation')


# class ShowNegateWarningTests(unittest.TestCase):

#     def _call(self, node, zodbnode):
#         report = IdentificationReport(None, None)
#         return report.show_negate_warning(node, zodbnode)

#     def test_show_Unanswered(self):
#         # https//code.simplon.biz/tracker/tno-euphorie/ticket/75
#         zodbnode = Risk()
#         zodbnode.problem_description = None
#         node = model.Risk(type="risk")
#         self.assertEqual(self._call(node, zodbnode), False)

#     def test_RiskNotPresent(self):
#         zodbnode = Risk()
#         zodbnode.problem_description = None
#         node = model.Risk(type="risk", identification="yes")
#         self.assertEqual(self._call(node, zodbnode), False)

#     def test_RiskNotApplicable(self):
#         zodbnode = Risk()
#         zodbnode.problem_description = None
#         node = model.Risk(type="risk", identification="n/a")
#         self.assertEqual(self._call(node, zodbnode), False)

#     def test_Present(self):
#         zodbnode = Risk()
#         zodbnode.problem_description = None
#         node = model.Risk(type="risk", identification="no")
#         self.assertEqual(self._call(node, zodbnode), True)

#     def test_HasProblemDescription(self):
#         zodbnode = Risk()
#         zodbnode.problem_description = u"Negative"
#         node = model.Risk(type="risk", identification="no")
#         self.assertEqual(self._call(node, zodbnode), False)

#     def test_HasEmptyProblemDescription(self):
#         zodbnode = Risk()
#         zodbnode.problem_description = u"   "
#         node = model.Risk(type="risk", identification="no")
#         self.assertEqual(self._call(node, zodbnode), True)


# XXX Change these test to check client.docx.html._HtmlToWord instead

# class HtmlToRtfTests(unittest.TestCase):

#     def HtmlToRtf(self, *a, **kw):
#         return HtmlToRtf(*a, **kw)

#     def render(self, output):
#         document = Document()
#         section = Section()
#         for o in output:
#             section.append(o)
#         document.Sections.append(section)
#         renderer = Renderer()
#         renderer.Write(document, StringIO())  # Setup instance variables
#         renderer._doc = document
#         renderer._fout = StringIO()
#         renderer._CurrentStyle = ""
#         renderer._WriteSection(section, True, False)
#         return renderer._fout.getvalue()

#     def testEmptyInput(self):
#         self.assertEqual(self.HtmlToRtf(u"", u"<stylesheet>"), [])

#     def testInvalidHtmlFallback(self):
#         self.assertTrue(
#             "text\\par" in
#             self.render(self.HtmlToRtf(u"<p>text</p>", u"<stylesheet>"))
#         )

#     def testBasicParagraph(self):
#         self.assertTrue(
#             "Simple text\\par" in self.render(
#                 self.HtmlToRtf(u"<p>Simple text</p>", u"<stylesheet>")
#             ), []
#         )

#     def testItalicInText(self):
#         self.assertTrue(
#             "Simple {\\i text}\\par" in self.render(
#                 self.
#                 HtmlToRtf(u"<p>Simple <em>text</em></p>", u"<stylesheet>")
#             )
#         )

#     def testBoldAndItalicText(self):
#         self.assertTrue(
#             "Very {\\i very }{\\b\\i bold}\\par" in self.render(
#                 self.HtmlToRtf(
#                     u"<p>Very <em>very <strong>bold</strong></em></p>",
#                     u"<stylesheet>"
#                 )
#             )
#         )

#     def testEmphasisInText(self):
#         self.assertTrue(
#             "{\\i text}" in
#             self.render(self.HtmlToRtf(u"<em>text</em>", u"<stylesheet>"))
#         )

#     def testInlineEntity(self):
#         self.assertTrue(
#             "Simple & clean\\par" in self.render(
#                 self.HtmlToRtf(u"<p>Simple &amp; clean</p>", u"<stylesheet>")
#             )
#         )

#     def testInlineEntityDigit(self):
#         self.assertTrue(
#             "Simple \r clean\\par" in self.render(
#                 self.HtmlToRtf(u"<p>Simple &#13; clean</p>", u"<stylesheet>")
#             )
#         )

#     def test_link_in_text(self):
#         # This demonstrates TNO Euphorie ticket 186
#         html = '<p>Check the <a rel="nofollow">manual</a> for more info.</p>'
#         rendering = self.render(self.HtmlToRtf(html, '<stylesheet>'))
#         self.assertTrue('Check the manual for more info.' in rendering)
#         self.assertEqual(rendering.count('more info'), 1)


class ActionPlanTimelineTests(EuphorieIntegrationTestCase):
    def setUp(self):
        super().setUp()
        self.account = addAccount(password="secret")

    def _get_timeline(self, context=None, request=None):
        """Return the timeline view."""

        class DummySurvey(mock.Mock, Base):
            __new__ = object.__new__

            def getPhysicalPath(self):
                return ("test", "dummy-survey")

        if request is None:
            request = self.request.clone()
            alsoProvides(request, IClientSkinLayer)
        if context is None:
            survey = DummySurvey()
            session = self._create_session()
            context = TraversedSurveySession(
                survey,
                session.id,
            ).__of__(survey)
        return api.content.get_view("timeline", context, request)

    def _create_session(self, dbsession=None):
        if dbsession is None:
            dbsession = Session()
        session = SurveySession(account=self.account, zodb_path="test/dummy-survey")
        dbsession.add(session)
        dbsession.flush()
        return session

    def test_get_measures_with_correct_module(self):
        view = self._get_timeline()
        session = view.context.session

        # This first module should be ignored, it doesn't contain any risks
        session.addChild(
            model.Module(
                zodb_path="1",
                module_id="1",
            )
        )
        # Between the next two modules, the first one (root-level) must be
        # returned.
        module = session.addChild(
            model.Module(
                zodb_path="2",
                module_id="2",
            )
        )
        module = module.addChild(
            model.Module(
                zodb_path="2/3",
                module_id="3",
            )
        )
        module.addChild(model.Risk(zodb_path="2/3/4", risk_id="1", identification="no"))
        survey = view.context.aq_parent
        survey.restrictedTraverse = lambda x, y: object
        survey.ProfileQuestions = lambda: []

        measures = view.get_measures()
        self.assertEqual(len(measures), 1)
        self.assertEqual(measures[0][0].module_id, "2")

    def test_get_measures_return_risks_without_measures(self):
        view = self._get_timeline()
        session = view.context.session

        module = session.addChild(
            model.Module(
                session=session,
                zodb_path="1",
                module_id="1",
            )
        )
        module.addChild(
            model.Risk(
                session=session, zodb_path="1/2", risk_id="1", identification="no"
            )
        )
        survey = view.context.aq_parent
        survey.restrictedTraverse = lambda x, y: object
        survey.ProfileQuestions = lambda: []

        measures = view.get_measures()
        self.assertEqual(len(measures), 1)
        self.assertEqual(measures[0][2], None)

    def test_get_measures_filter_on_session(self):
        view = self._get_timeline()
        sessions = [
            view.context.session,
            self._create_session(),
        ]
        for session in sessions:
            module = session.addChild(
                model.Module(
                    session=session,
                    zodb_path="1",
                    module_id="1",
                )
            )
            module.addChild(
                model.Risk(
                    session=session,
                    zodb_path="1/2",
                    risk_id="1",
                    identification="no",
                    action_plans=[
                        model.ActionPlan(
                            action="Measure 1 for %s" % session.account.loginname
                        )
                    ],
                )
            )

        survey = view.context.aq_parent
        survey.restrictedTraverse = lambda x, y: object
        survey.ProfileQuestions = lambda: []

        measures = view.get_measures()
        self.assertEqual(len(measures), 1)
        self.assertEqual(measures[0][2].action, "Measure 1 for jane@example.com")

    def test_get_measures_order_by_start_date(self):
        view = self._get_timeline()
        session = view.context.session
        module = session.addChild(
            model.Module(
                session=session,
                zodb_path="1",
                module_id="1",
            )
        )
        module.addChild(
            model.Risk(
                session=session,
                zodb_path="1/2",
                risk_id="1",
                identification="no",
                action_plans=[
                    model.ActionPlan(
                        action="Plan 2", planning_start=datetime.date(2011, 12, 15)
                    ),
                    model.ActionPlan(
                        action="Plan 1", planning_start=datetime.date(2011, 11, 15)
                    ),
                ],
            )
        )

        survey = view.context.aq_parent
        survey.restrictedTraverse = lambda x, y: object
        survey.ProfileQuestions = lambda: []

        measures = view.get_measures()
        self.assertEqual(len(measures), 2)
        self.assertEqual([row[2].action for row in measures], ["Plan 1", "Plan 2"])

    def test_priority_name_known_priority(self):
        view = self._get_timeline()
        self.assertEqual(view.priority_name("high"), "High")

    def test_priority_name_known_unpriority(self):
        view = self._get_timeline()
        self.assertEqual(view.priority_name("dummy"), "dummy")

    def test_create_workbook_empty_session(self):
        # If there are no risks only the header row should be generated.
        view = self._get_timeline()
        view.getModulePaths = lambda: []
        book = view.create_workbook()
        self.assertEqual(len(book.worksheets), 1)
        sheet = book.worksheets[0]
        self.assertEqual(len(tuple(sheet.rows)), 1)

    def test_create_workbook_plan_information(self):
        view = self._get_timeline()
        module = model.Module(
            zodb_path="1",
            title="Top-level Module title",
        )
        risk = model.Risk(
            zodb_path="1/2/3",
            risk_id="1",
            title="Risk title",
            priority="high",
            identification="no",
            path="001002003",
            comment="Risk comment",
        )
        plan = model.ActionPlan(
            action="Plan 2", planning_start=datetime.date(2011, 12, 15), budget=500
        )
        survey = view.context.aq_parent
        zodb_node = mock.Mock()
        zodb_node.problem_description = "This is wrong."
        survey.restrictedTraverse.return_value = zodb_node

        view.get_measures = lambda: [(module, risk, plan)]
        wb = view.create_workbook()
        sheet = wb.worksheets[0]
        # planning start
        self.assertEqual(sheet["A2"].value, datetime.date(2011, 12, 15))
        # planning end
        self.assertEqual(sheet["B2"].value, None)
        # action plan
        self.assertEqual(sheet["C2"].value, "Plan 2")
        # requirements
        self.assertEqual(sheet["D2"].value, None)
        # responsible
        self.assertEqual(sheet["E2"].value, None)
        # budget
        self.assertEqual(sheet["F2"].value, 500)
        # module title
        self.assertEqual(sheet["G2"].value, "Top-level Module title")
        # risk number
        self.assertEqual(sheet["H2"].value, "1.2.3")
        # risk title
        self.assertEqual(sheet["I2"].value, "This is wrong.")
        # risk priority
        self.assertEqual(sheet["J2"].value, "High")
        # risk comment
        self.assertEqual(sheet["K2"].value, "Risk comment")

    def test_create_workbook_no_problem_description(self):
        view = self._get_timeline()
        module = model.Module(
            zodb_path="1",
            path="001",
            title="Top-level Module title",
        )
        risk = model.Risk(
            zodb_path="1/2/3",
            risk_id="1",
            title="Risk title",
            priority="high",
            identification="no",
            path="001002003",
            comment="Risk comment",
        )
        survey = view.context.aq_parent
        survey.ProfileQuestions = lambda: []
        zodb_node = mock.Mock()
        zodb_node.title = "Risk title."
        zodb_node.problem_description = "  "
        survey.restrictedTraverse.return_value = zodb_node
        view.getRisks = lambda x: [(module, risk)]
        sheet = view.create_workbook().worksheets[0]
        self.assertEqual(sheet["I2"].value, "Risk title")

    def test_render_value(self):
        with api.env.adopt_user(user=self.account):
            view = self._get_timeline()
            view.context.session.title = "Acmè"
            survey = view.context.aq_parent
            survey.ProfileQuestions = lambda: []
            view.__call__()
            response = view.request.response
            self.assertEqual(
                response.headers["content-type"],
                "application/vnd.openxmlformats-" "officedocument.spreadsheetml.sheet",
            )
            quoted_filename = quote("Action plan for Acmè.xlsx".encode())
            self.assertEqual(quoted_filename, "Action%20plan%20for%20Acm%C3%A8.xlsx")
            self.assertEqual(
                response.headers["content-disposition"],
                f"attachment; filename*=UTF-8''{quoted_filename}",
            )


class UnansweredQueryTests(EuphorieIntegrationTestCase):
    """Test #2800.

    Only risks with `identification` == None are returned as unanswered.
    They are not returned as not present.
    """

    def createData(self):
        (self.session, self.survey_session) = createSurvey()

        self.q1 = model.Module(
            **{
                "depth": 1,
                "module_id": 1,
                "has_description": True,
                "path": "001",
                "postponed": None,
                "profile_index": -1,
                "skip_children": False,
                "title": "What is the sound of one hand clapping?",
                "type": "module",
                "zodb_path": "173",
            }
        )
        self.survey_session.addChild(self.q1)

        self.mod1 = model.Module(
            **{
                "depth": 2,
                "module_id": 2,
                "has_description": True,
                "path": "001001",
                "postponed": None,
                "profile_index": 0,
                "skip_children": False,
                "title": "Stellenbosch",
                "type": "module",
                "zodb_path": "173",
            }
        )
        self.q1.addChild(self.mod1)

        self.r1 = model.Risk(
            **{
                "risk_id": 1,
                "depth": 3,
                "identification": "no",
                "action_plans": [],
                "has_description": True,
                "path": "001001001",
                "postponed": False,
                "profile_index": 0,
                "skip_children": False,
                "title": "Hands are washed",
                "type": "risk",
                "zodb_path": "173/euphorie.risk",
            }
        )
        self.mod1.addChild(self.r1)

        self.r3 = model.Risk(
            **{
                "risk_id": 2,
                "depth": 3,
                "identification": "yes",
                "action_plans": [],
                "has_description": True,
                "path": "001001002",
                "postponed": False,
                "profile_index": 0,
                "skip_children": False,
                "title": "Feet are washed",
                "type": "risk",
                "zodb_path": "173/euphorie.risk-1",
            }
        )
        self.mod1.addChild(self.r3)

        self.mod2 = model.Module(
            **{
                "depth": 2,
                "module_id": 3,
                "has_description": True,
                "path": "001002",
                "postponed": None,
                "profile_index": 1,
                "skip_children": False,
                "title": "Somerset West",
                "type": "module",
                "zodb_path": "173",
            }
        )
        self.q1.addChild(self.mod2)

        self.r2 = model.Risk(
            **{
                "risk_id": 1,
                "depth": 3,
                "identification": None,
                "action_plans": [],
                "has_description": True,
                "path": "001002001",
                "postponed": False,
                "profile_index": 1,
                "skip_children": False,
                "title": "Hands are washed",
                "type": "risk",
                "zodb_path": "173/euphorie.risk",
            }
        )
        self.mod2.addChild(self.r2)

    def testUnansweredNodes(self):
        self.createData()
        risks = [
            node
            for node in utils.get_unanswered_nodes(self.survey_session)
            if node.type == "risk"
        ]
        self.assertEqual(len(risks), 1)

    def testRiskNotPresentNodes(self):
        self.createData()
        risks = [
            node
            for node in utils.get_risk_not_present_nodes(self.survey_session)
            if node.type == "risk"
        ]
        self.assertEqual(len(risks), 1)



from Acquisition import aq_parent
from euphorie.content.tests.utils import createSector
from euphorie.testing import EuphorieIntegrationTestCase


class CountryTests(EuphorieIntegrationTestCase):
    def createCountry(self):
        sector = createSector(self.portal, login="sector")
        return aq_parent(sector)

    def testCanNotBeCopied(self):
        country = self.createCountry()
        self.assertFalse(country.cb_isCopyable())

    def testDefaultInfoSections(self):
        country = self.createCountry()
        self.assertEqual(
            country.risk_default_collapsible_sections,
            ["collapsible_section_information"],
        )

    def testDefaultReports(self):
        country = self.createCountry()
        self.assertEqual(
            country.default_reports,
            ["report_full", "report_action_plan", "report_overview_risks"],
        )

    def testDefaultTraining(self):
        country = self.createCountry()
        self.assertEqual(
            country.enable_web_training,
            True,
        )



from euphorie.client import model
from euphorie.client.profile import always_present_default
from euphorie.client.utils import HasText
from euphorie.content.interfaces import IQuestionContainer
from z3c.saconfig import Session

import collections


def getSurveyTree(survey, profile=None):
    """Return a list of all modules, profile questions and risks in a survey.

    Each entry is represented by a dict with a `zodb_path` and `type`
    key.
    """
    # XXX Can this be cached on the survey instance?
    nodes = []
    base_length = len(survey.getPhysicalPath())
    queue = collections.deque(survey.values())
    while queue:
        node = queue.popleft()
        if node.portal_type not in [
            "euphorie.profilequestion",
            "euphorie.module",
            "euphorie.risk",
        ]:
            continue
        # Note that in profile.AddToTree, we pretend that an optional module
        # always has a description. This logic needs to be replicated here.
        if node.portal_type == "euphorie.module":
            has_description = HasText(node.description) or node.optional
        else:
            has_description = HasText(node.description)
        if profile and node.portal_type == "euphorie.profilequestion":
            if not profile.get(node.id):
                continue
        nodes.append(
            {
                "zodb_path": "/".join(node.getPhysicalPath()[base_length:]),
                "type": node.portal_type[9:],
                "has_description": has_description,
                "always_present": (
                    node.portal_type[9:] == "risk" and node.risk_always_present or False
                ),
                "risk_type": (node.portal_type[9:] == "risk" and node.type or None),
                "optional": node.optional,
            }
        )
        if IQuestionContainer.providedBy(node):
            queue.extend(node.values())
    return nodes


class Node:
    """Utility class to store information for a survey tree node.

    This is a subset of the data in
    :py:class:`euphorie.client.model.SurveyTreeItem` and is hashable so
    it can be stored in a `set`.
    """

    def __init__(self, item):
        self.zodb_path = item.zodb_path
        self.path = item.path
        self.type = item.type
        self.skip_children = item.skip_children
        self.has_description = item.has_description
        self.identification = item.type == "risk" and item.identification or None
        self.risk_type = item.type == "risk" and item.risk_type or None

    def __repr__(self):
        return "<Node zodb_path={} type={} path={}>".format(
            self.zodb_path,
            self.type,
            self.path,
        )

    def __hash__(self):
        return hash(self.path)


def getSessionTree(session):
    """Build and return a list of all survey tree nodes (as :obj:`Node`
    instances) for the current session."""
    sqlsession = Session()
    query = (
        sqlsession.query(model.SurveyTreeItem)
        .filter(model.SurveyTreeItem.session_id == session.id)
        .order_by(model.SurveyTreeItem.path)
    )
    return [Node(item) for item in query]


def treeChanges(session, survey, profile=None):
    """Determine a list of changes in a survey for an existing session.

    The list of changes is returned as a list of tuples listing the ZODB
    path, the object type (one of `profile`, `module` or `risk`) and the
    change type (one of `add`, `remove` or 'modified').
    """
    surveytree = getSurveyTree(survey, profile)
    sestree = set(getSessionTree(session))
    results = set()
    for entry in surveytree:
        nodes = [node for node in sestree if node.zodb_path == entry["zodb_path"]]
        if nodes:
            for node in nodes:
                sestree.remove(node)

            if nodes[0].type == entry["type"] == "module":
                if entry["optional"] is False and nodes[0].skip_children is True:
                    # skip_children cannot be True if the module is not
                    # optional, so this is requires a SessionTree update
                    results.add((entry["zodb_path"], nodes[0].type, "modified"))
                elif entry["has_description"] != nodes[0].has_description:
                    # Log module description changes since this changes
                    # client behaviour: modules without description are
                    # skipped.
                    results.add((entry["zodb_path"], nodes[0].type, "modified"))
            if node.type == entry["type"] == "risk":
                if (
                    entry["always_present"]
                    and node.identification != always_present_default
                ):
                    results.add((entry["zodb_path"], node.type, "modified"))
                if entry["risk_type"] != node.risk_type:
                    results.add((entry["zodb_path"], node.type, "modified"))
            if nodes[0].type == entry["type"] or (
                nodes[0].type == "module" and entry["type"] == "profilequestion"
            ):
                continue
            # Flag a type change as remove & add
            results.add((entry["zodb_path"], nodes[0].type, "remove"))
            results.add((entry["zodb_path"], entry["type"], "add"))
        else:
            results.add((entry["zodb_path"], entry["type"], "add"))

    for node in sestree:
        if node.zodb_path.find("custom-risks") != -1:
            continue
        results.add((node.zodb_path, node.type, "remove"))

    return results


def wasSurveyUpdated(session, survey):
    """Check if a survey was updated (ie republished) after the last
    modification in a session.

    If the survey was updated but no changes were made in its structure,
    for example if there were only textual changes, this method updates
    the timestap on the survey session and pretends there was no update.
    """
    # BBB: Old surveys had no published timestamp. Can be removed post
    # site launch.
    published = getattr(survey, "published", None)
    if published is not None:
        if isinstance(published, tuple):
            timestamp = published[2]
        else:
            # BBB: Euphorie 1.x did not use a tuple to store extra information.
            timestamp = published
        if session.refreshed >= timestamp:
            return False

    changes = treeChanges(session, survey)
    if not changes:
        session.refresh_survey(survey)
        return False

    return True
